#!/usr/bin/python3

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import time
import sys
from xarray import DataArray

from caproto.sync import client as ca_client
#from caproto.sync import server as ca_server

from prometheus_client import start_http_server, Summary, Counter, Gauge

import argparse
import logging
import math
import xarray

from emmi.ca.reader import GuidedPvReader, PvRetry

class LiveDisplay:
    '''
    Uses Matplotlib to visualize numpy arrays (one or several?)
    '''
    def __init__(self, panels=None):
        '''
        Intialises a Matplotlib figure with the named panels.
        
        Parameters:
          - `panels`: A string array with panel reference names,
            or optionally a dictionary with reference and a displayable
            description.
        '''

        self.panelNames = panels or { "default": "Default display" }

        rows = int(round(math.sqrt(len(self.panelNames))))
        cols = int(math.ceil(len(self.panelNames)/rows))

        logging.info("Starting plotting interface")

        # We have a figure with NxM subplots (i.e. axes), each subplot hosting
        # one lineplot (for now), each referenced by name.
        self.figure = plt.figure()
        self.axes = { k: self.figure.add_subplot(rows, cols, i+1) for i,k in enumerate(self.panelNames) }
        self.lines = { k: x.plot([])[0] for k, x in zip(self.panelNames, self.figure.axes) }

        # caching some data array metrics for autoscaling
        self.last_update = {k:time.time() for k in self.panelNames }
        
        self.figure.show(True)
        

    def update(self, panel, data=None, xaxis=None, xlabel=None, ylabel=None, text=None):
        '''
        Updates data on the panel named `panel` using the data from `data`.
        `data` should preferrably be an `xarray.DataArray` container, in
        which case all metadata will be used extracted from there.
        Otherwise axis scalind and labeling will be used from the named
        arguments.

        The panels will upscale the X and Y ranges to fit all the data
        dynamics, but will never scale down.
        '''

        try:
            ax = self.axes[panel]
        except KeyError:
            logging.error("%s: no such display panel" % panel)
            return

        if xlabel:
            ax.set_xlabel(xlabel)
            
        if ylabel:
            ax.set_ylabel(ylabel)

        if data is None:
            return
        
        #if len(data) != self.xlen[panel]:
        #    self.xlen[panel] = len(data)
        #xlen = len(data)

        if xaxis is None:
            xaxis = data.coords[data.dims[0]].values if isinstance(data, xarray.DataArray) \
                else np.array([i for i in range(len(data))])

        xlim = np.array([xaxis[0], xaxis[-1]])
        if (xlim != ax.get_xlim()).all():
            logging.info("Adjusting X axis to %r" % xlim)
            ax.set_xlim(*xlim)

        ylim = ax.get_ylim()
        dlim = np.array([data.min(), data.max()])
        #print(data)
        if (dlim[0] < ylim[0]) or (dlim[1] > ylim[1]):
            logging.info("Adjusting Y range to %r" % dlim)
            ax.set_ylim(min(dlim[0], ylim[0]), max(dlim[1], ylim[1]))

        self.lines[panel].set_data(xaxis, data)
        self.figure.canvas.draw_idle()

        # just for benchmarking.
        tnow = time.time()
        tdiff = tnow - self.last_update[panel]
        self.last_update[panel] = time.time()

        #text.set_text("acquisition: %2.2f Hz | flags: %r | counts: %d" % \
        #              (1.0/(tnow-t0),
        #               [], # f for f in hdev.flags
        #               data.sum()))

        
    def handle_events(self):
        self.figure.canvas.flush_events()
        
        

def old_histo_loop(fig, aqtime=0.1, axis_index=0, line_index=0, channel=0):
    '''
    Loops endlessly reading histogram and displaying it into the
    specified Matplotlib plot, while also serving the GUI event loop.
    '''
    
    print("Starting plotting interface")
    ax = fig.axes[axis_index] if len(fig.axes)>0 else fig.add_subplot(1, 1, 1)
    line = ax.lines[line_index] if len(ax.lines)>0 else ax.plot([])[0]
    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Intensity (counts)')
    text  = ax.text(0, 0.0, "Jinkies")
    fig.show(False)

    devPrefix = "KMC3:XPP:HARP:"
    chPrefix = devPrefix+"CH%d_" % channel

    #ca.write(devPrefix+"ACQUISITIONTIME", 0.5)
    
    # For auto-scaling
    xlen = 0
    ymax = 0
    t0 = time.time()

    while True:

        # wait for acquisition to end
        while True:
            try:
                v = ca_client.read(devPrefix+"ACQUIRINGRBV")
                #print("Waiting for acquisition to end... (%r)" % v.data)
                if v.data[0].decode('utf-8') == "OFF":
                    break
            except Exception as e:
                print("Error:", e)
            fig.canvas.flush_events()
            time.sleep(0.1)

        try:
            data = np.array(ca_client.read(chPrefix+"HISTOGRAM_SIGNAL").data)
            offs = ca_client.read(chPrefix+"HISTOGRAM_OFFSET").data[0]
            delta = ca_client.read(chPrefix+"HISTOGRAM_DELTA").data[0]
            reach = offs+delta*len(data)
            xaxis = np.array([offs+delta*i for i in range(len(data))])
            #xdata = DataArray(data=data, dims=["delay"], coords=[xaxis])
        except Exception as e:
            print("Error:", e)

        print ("Acquisition finished:", data.shape, data, offs, delta, data[[i!=0 for i in data]])

            
        if len(data) != xlen:
            xlen = len(data)
            ax.set_xlim(offs, reach)

        ## Some random noise, for testing
        #data[np.random.randint(xlen)] = np.random.randint(100)
        
        dmax = data.max()
        if ymax < dmax:
            print ("Adjusting Y range")
            ymax = dmax
            ax.set_ylim(-1, dmax+1)

        tnow = time.time()
        text.set_text("acquisition: %2.2f Hz | flags: %r | counts: %d" % \
                      (1.0/(tnow-t0),
                       [], # f for f in hdev.flags
                       data.sum()))
        t0 = tnow

        print ("Updating plot (heartbeat: %r, max %d @ %d)" % \
               (ca_client.read(devPrefix+"heartbeat").data[0], dmax, -1))
        line.set_data(xaxis, data)
        fig.canvas.draw_idle()
        fig.canvas.flush_events()

        # This turns on the acquisition
        #ca.write(devPrefix+"ACQUIRINGVAL", "ON")
        
        while True:
            try:
                v = ca_client.read(devPrefix+"ACQUIRINGRBV").data
                #print("Waiting for acquisiton to start... (%r)" % v)
                if v.data[0].decode('utf-8') == "ON":
                    break
            except Exception as e:
                print("Error:", e)
            fig.canvas.flush_events()
            time.sleep(0.1)

        print ("Acquiring...")
    

class FifoArrayWrapper:
    '''
    Quick 'n Dirty wrapper 
    '''
    def __init__(self, length, *args, **kw):
        self.data = np.zeros(length)

    def push_value(self, val):
        current = (self.data[1:]).copy()
        self.data[:-1] = current[:]
        self.data[-1] = val


def save_me(dataMap):
    pass

def save_array(path, data):
    print ("Saving:", path)
    with open(path, "w") as f:
        f.write(" ".join([str(i) for i in data]))


def histo_loop(display, reader=None):
    '''
    This is where the magic happens.
    '''
    
    countRate    = FifoArrayWrapper(512)
    #roi1         = FifoArrayWrapper(512)
    #roi2         = FifoArrayWrapper(512)
    #roi3         = FifoArrayWrapper(512)
    #roi4         = FifoArrayWrapper(512)
    #ratio_roi12  = FifoArrayWrapper(512) 

    # about xarray:
    
    hstack = None
    
    while True:
        try:
            dataMap = reader.retr()

            #save_me(dataMap)

            print("Have:", dataMap)

            countRate.push_value(dataMap['CH0_COUNTRATE'])
            
            display.update('histo', dataMap['CH0_HISTOGRAM_SIGNAL'], xlabel='Time (s)', ylabel='Counts')

            # Saving data
            #filebase = '/tmp/histo'
            #tstamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())+("%.2f"%(time.time()%1))[1:]  
            #save_array(filebase+"-"+tstamp, dataMap['CH0_HISTOGRAM_SIGNAL'].values)
            
            #if hstack is None:
            #	hstack = dataMap['CH0_HISTOGRAM_SIGNAL'].copy()
            #else:
            #	hstack += dataMap['CH0_HISTOGRAM_SIGNAL']
            #    
            #display.update('hstack', hstack, xlabel='Time (s)', ylabel='Counts')
            
            # Check out documentation for 'xarray' on indexing these containers
            #roi1_data = hstack.sel(x=slice(1e-11, 5e-8))
            #roi2_data = hstack.sel(x=slice(1e-7, 1e-6))
            #roi3_data = hstack.sel(x=slice(1e-6, 1e-5))
            #roi4_data = hstack.sel(x=slice(1e-8, 1e-7))
            #sum_data = hstack.sel(x=slice(0, 2e-6))
            #print(sum_data.sum())
            #print(np.any(np.isnan(dataMap['CH0_HISTOGRAM_SIGNAL'].values)))
            
            #roi1.push_value(roi1_data.sum())
            #roi2.push_value(roi2_data.sum())
            #roi3.push_value(roi3_data.sum())
            #roi4.push_value(roi4_data.sum())
            #if roi2_data.sum() == 0:
            #    ratio_roi12.push_value(0)
            #else:
            #    ratio_roi12.push_value((roi2_data.sum()-roi1_data.sum())/roi2_data.sum())
            
            #ca_client.write("KMC3:XPP:HARP:ACQUIRINGVAL", "ON")
            
            display.update('rate', countRate.data, ylabel='Count rate')
            #display.update('roi1', roi1.data, ylabel='ROI 1 (sum counts)')
            #display.update('roi2', roi2.data, ylabel='ROI 2 (sum counts)')
            #display.update('roi3', roi3.data, ylabel='ROI 3 (sum counts)')
            #display.update('roi4', roi4.data, ylabel='ROI 4 (sum counts)')
            #display.update('ratio_roi12', ratio_roi12.data, ylabel='ratio (ROI2-ROI1)/ROI2')
            
        except PvRetry:
            time.sleep(0.05)
            
        except Exception as e:
            print(e)
            time.sleep(3)
            
        except:
            logging.error("You broke it.")
            time.sleep(10)
            
        display.handle_events()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(prog="capeek", description="EPICS Channel Access data peek utility")
    parser.add_argument('-l', '--loglevel', action='store', default='INFO')
    parser.add_argument('--legacy', action='store_true', default=False)
    parser.add_argument('-p', '--mport', action='store', default=31415,
                        help='Port where to export summaries of data for Prometheus monitoring')
    parser.add_argument('-P', '--ioc-prefix', action='store', default='',
                        help='Prefix to use for all IOC variables')
    parser.add_argument('-H', '--ioc-heartbeat', action='store', default='heartbeat',
                        help='Heartbeat variable suffix')

    opts = parser.parse_args()

    level = getattr(logging, (opts.loglevel or 'INFO').upper(), None)
    logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", level=level)
    
    heartbeat = Counter((opts.ioc_prefix+opts.ioc_heartbeat).lower().replace(':', '_'),
                        'IOC master heartbeat for %s' % opts.ioc_prefix)

    start_http_server(opts.mport)

    if opts.legacy:
        logging.info("Starting legacy display")
        old_histo_loop(plt.figure())
        
    else:
        logging.info("New and improved, but broken, display")

        reader = GuidedPvReader(prefix="KMC3:XPP:HARP:",
                                guides={ "ACQUIRINGRBV": "OFF" },
                                pv=["CH0_HISTOGRAM_SIGNAL",
                                    "CH0_COUNTRATE"])
        
        displ = LiveDisplay(panels=['histo', 'rate'])
        
        histo_loop(displ, reader)



#!/usr/bin/python3

import pytest
import time
from emmi.eda import MotorEngine, MockMotor

import sys

def test_mock_motor():
    tslice = 1.0
    motor = MockMotor(mock_timeslice=tslice, limits={'lo': -3.0, 'hi': +2.0})

    # initial state should be "stopped"
    assert motor.moves() == False

    # moving by dynamic speed, always takes tslice seconds,
    # and .moves() is True during the motion phase
    motor.goto(-2.0)
    assert motor.moves()
    t0 = time.time()
    while motor.moves():
        time.sleep(0.01)
    td = time.time()-t0

    assert motor.moves() == False
    assert td > tslice*0.8
    assert td < tslice*1.2

    # motor stops at lower limit and signals flag.
    motor.goto(-4.0)
    while motor.moves():
        time.sleep(0.01)
    
    assert "lo" in motor.flags
    assert motor.where() < -2.9
    assert motor.where() > -4.1

def test_engine():
    '''
    Very rudimentary test of the engine, using a Mock motor:
      - Initial state is INIT or IDLE
      - After setting position we're BUSY for a while, then IDLE
      - After setting and stopping we're going back to IDLE
    '''
   
    e = MotorEngine(motor=MockMotor(mock_timeslice=1.0))
    assert e.step() in [ "INIT", "IDLE" ]

    # Setting the position.
    # Will need 1 second to move, may or may not need 1 second to stop (?)
    e.go(3.14)
    end = time.time()+2.5
    while time.time() < end:
        time.sleep(0.1)
        assert e.step() in [ "STAGING", "BUSY", "IDLE", "STOP" ]

    e.go(10.0)
    assert e.step() in [ "BUSY", "STAGING" ]
    
    e.stop()
    assert e.step() == "STOP"

    end = time.time()+1.5
    while time.time() < end:
        time.sleep(0.1)
        s = e.step()
        
    assert e.step() == "IDLE"

    ## Also need to test that setting a position while already
    ## moving results in ignoring the command.




#!/usr/bin/python3

import time
import asyncio
import logging

from functools import partial

logger = logging.getLogger(__name__)

class MockMotor(object):
    '''
    The simplest of the simplest motors: can only move, report position and stop.

    Also exports `.errors` and `.flags`. The first is expected to be an empty
    list when the motor state is "ok" (i.e. no error), or a list with one
    or several objects (e.g. strings) when errors with the motor persit.
    Calling `.clear()` clears the error list.

    The `.flags` member is expected to be an enumerable data type (list,
    tuple, dictionary, set or something similar) to convey supplementary
    information about motor events, e.g. triggering of low / high switch
    limits etc.

    With the exception of limit switches (see `.__init__()`), `MockMotor`
    does not intrinsically handle or simulate any other switches.
    '''

    def __init__(self, *args, **kwargs):
        ''' Creates a mock-up of an EDA Motor.

        Accepts all args as a courtesy of being a good mock-up class,
        but ignores most of them. A number of arguments are specific
        to the mock class. They are listed below.

        Args:
            mock_timeslice: a measure of how "fast" a mock motor is moving.
              Every `.goto()` call on a mock motor will execute a movement
              of the specified distance in exactly `mock_timeslice` seconds,
              regardless of the distance.

            limits: tuple of `(low, high)` to simulate the lower,
              respectively higher hardware limits. If any of those
              is reached in the mocking position, `.flags` will contain
              `"LLIM"` or `"HLIM"`, respectively. Alternatively, this can
              also be a dictionary with two or three items. If it's two
              items, it's used as the low/high limits, and the
              keys are used in `.flags` instead of `"HLIM"`/`"LLIM"`.
            
        '''
        self.mock_timeslice = kwargs.get('mock_timeslice', 5.0)
        self.start = 0.0
        self.target = 0.0
        self.tstamp = 0.0
        self.errors = []
        self._limits = kwargs.get('limits', None)
    
    def where(self):
        '''
        Returns current position -- that is the position that we'd be currently
        having if we'd wanted to go from "current" towards "target" within
        the timeslice "mock_timeslice"
        '''
        tdelta = (time.time()-self.tstamp)
        if tdelta > self.mock_timeslice:
            tdelta = self.mock_timeslice
        dist = self.target-self.start
        return self.start + dist * (tdelta/self.mock_timeslice)

    def goto(self, val):
        '''
        Sends command to move to position (doesn't wait)
        '''
        self.start = self.where()
        self.target = val
        self.tstamp = time.time()

    def stop(self):
        '''
        Sends command to stop (doesn't wait). Technically, we'll be still
        in state "moving" for a while after this, but we'd be moving
        towards the position we're already onto.
        '''
        if self.moves():
            self.goto(self.where())

    def moves(self):
        '''
        Returns True if the motor moves. We fake this by testing whether
        we're still within the "timeslice". This has the added benefit
        that sometimes moves() returns False right away (i.e. if we weren't
        moving in the first place), and sometimes still returns False
        for a considerate amount of time (i.e. until the end of the current
        slice) if we were moving and just received a stop() command.
        '''
        now = time.time()
        return (now - self.tstamp) <= self.mock_timeslice

    def clear(self):
        '''
        Clears all outstanding error flags (they may pop up again).
        '''
        self.errors = []

    @property
    def flags(self):
        ''' Check for HLIM / LLIM and return the appropriate strings.

        Strings are either default "HLIM", "LLIM" respectively, or
        the keys of the `._limits` parameter.
        '''
        if self._limits is None:
            return set()

        f = set()

        lk = [k for k in self._limits.keys()]

        low = ('LLIM', self._limits[0]) if not isinstance(self._limits, dict) \
            else (lk[0], self._limits[lk[0]])

        high = ('HLIM', self._limits[-1]) if not isinstance(self._limits, dict) \
            else (lk[-1], self._limits[lk[-1]])

        p = self.where()
        
        if p <= low[1]:
            f.add(low[0])

        if p >= high[1]:
            f.add(high[0])

        return f
    

class MotorEngine(object):
    '''
    Generic motor engine class. Essentially cycles through the states
    as it should, and serves the most basic properties `position` and
    `position_relative` correctly. It's based on generic True/False
    conditional procedures to decide when to switch.

    In its default incarnation, the conditionals are just generic waiting
    functions (1 second for anything), so it can be used as a mock-up for
    unit testing. But if you replace the conditionals by things that do
    "real stuff", it's actually usable for production.
    '''
    
    def __init__(self, motor=None, init_probe=None):
        '''
        Initializes a motor engine.

        Args:
            motor: a `MockMotor` compatible motor hardware driver
        
            init_probe: if not `None`, it's a callable of the form `func(motor)`,
              which should return `False` for as long as the motor is still
              initializing.)
        '''
        
        self.__motor = motor or MockMotor()

        # current state
        self.__state = "INIT"
        self.__old_state = None

        self.__init_probe = init_probe

        # sub-type of the "BUSY" state
        self.__business = None

        # When this is not None, it is expected to hold a tuple
        # (name, go_callable, busy_callable) that triggers a new BUSY state.
        self.__scheduled_go = None

        # This is the similar to "__scheduled_go", only that the name is "STOP",
        # the go-callable is motor.stop(), and busy_check is motor.moves().
        # It's necessary to keep this separate from __scheduled_go because this
        # is actually supposed to be able to preemt __scheduled_go at any time.
        # This needs for special handling in the code.
        self.__scheduled_stop = None


    @property
    def business(self):
        if self.__state in [ "BUSY", "STAGING" ]:
            return self.__business
        else:
            return None
    

    @property
    def errors(self):
        return self.__motor.errors


    def state_INIT(self, state):
        # ...put init stuff here.
        if self.__init_probe is None:
            return "IDLE"
        if self.__init_probe(self.__motor):
            return "IDLE"
        
        return "INIT"


    # STOP state: trigger STOP hardware command, stay here while hardware is moving
    def state_enter_STOP(self, state):
        if self.__scheduled_stop:
            self.__scheduled_stop[1]()
        return state

    def state_STOP(self, state):
        if self.__scheduled_stop and self.__scheduled_stop[2]():
            return "STOP"
        self.__scheduled_go = None
        self.__scheduled_stop = None
        if len(self.errors) > 0:
            return "ERROR"
        return "IDLE"


    # BUSY state: trigger busy action, stay here while action not done
    def state_enter_STAGING(self, state):
        
        if self.__scheduled_stop: # STOP trumps everything
            logger.debug(f'STAGING: {self.__scheduled_go["name"]} requested, but also have a stop request pending')
            return "STOP"
        
        logger.debug(f'STAGING: {self.__scheduled_go["name"]}')
        self.__scheduled_go["proc"]()
        return state

    def state_STAGING(self, state):
        if self.__scheduled_go["busy"]() == True:
            return "BUSY"


    def state_BUSY(self, state):
        
        if len(self.errors) > 0:
            return "STOP"
        
        if self.__scheduled_stop is not None:
            return "STOP"
        
        if self.__scheduled_go["busy"]() == False:
            return "STOP"


    # IDLE state: stay here until there is work scheduled.
    def state_IDLE(self, state):
        
        if len(self.errors) > 0:
            logger.debug(f'Have {len(self.errors)}, STOP-ing motor')
            return "STOP"
        
        if self.__scheduled_go is not None:
            self.__business = self.__scheduled_go["name"] or "(default)"
            logger.debug(f'Going BUSY, business {self.__business}')
            return 'STAGING'


    def state_ERROR(self, state):
        '''
        The only way we can leave ERROR is by clearing/going to IDLE
        '''
        if len(self.errors) == 0:
            return "IDLE"


    # There's no escape from death.
    def state_FAIL(self, state):
        pass
             

    ## State procedures -- execute the current state procedure.
    ## If the state (just) switched, execute "state_enter_{state}" instead.    
    def state_proc(self):

        state = self.__state

        # BUSY state needs some special treatment -- it may be of the form 'BUSY.{}'
        s_proc_name = state if not state.startswith("BUSY.") else state[:4]
        s_proc = getattr(self, f'state_{s_proc_name}')

        #print(f"State run: {self.__old_state} {state}")

        # If we just entered the current state, maybe there's a dedicated state_enter_{...}
        if state != self.__old_state:
            logger.debug ("State: %s -> %s" % (self.__old_state, state))
            try:
                s_proc = getattr(self, f'state_enter_{state}')
            except AttributeError:
                pass

        new_state = s_proc(state)
            
        self.__old_state = state

        if new_state is not None:
            self.__state = new_state

        return self.__state


    async def run(self, period=0.01, stop_on_fail=True):
        '''
        Async function to run the Motor Engine
        '''
        if self.__state == "FAIL":
            raise RuntimeError("Attempted to start a motor engine in a failed state")

        while (self.__state not in [ "FAIL" ]) or (not stop_on_fail):
            try:
                self.state_proc()
                await asyncio.sleep(period)
            except Exception as e:
                logger.error(f'Motor engine failed: {str(e)}')
                self.__state == "FAIL"

        logger.error("Motor engine in FAIL state")

        
    @property
    def state(self):
        '''
        Returns the current state
        '''
        return self.__state

    
    def step(self):
        '''
        Returns the current state after having performed one state-step.
        '''
        return self.state_proc()
    
    
    # Current position -- getter reads out the hardware, setter
    # is a bit more complicated because we have to acknowledge
    # the current state (i.e. can't just override current states).
    @property
    def position(self):
        return self.__motor.where()
    
    @position.setter
    def position(self, val):
        self.go(val)

    
    def stop(self):
        '''
        Triggers an exit from any BUSY state into stop.
        '''        
        self.__scheduled_stop = ("stop",
                                 lambda: self.__motor.stop(),
                                 lambda: self.__motor.moves())
        logger.debug(f"Stop requested, current state is {self.state}")
        
        
    def go(self, call, *call_args,
           name=None,
           done_flag=None,
           busy_flag=None,
           busy_check=None,
           **call_kwargs):
        '''
        Triggers a work session / BUSY round.

        Args:
        
            call: work task. If it's a callable, it's called as it is.
              If it's a string, it is interpreted to be a member of the motor instance
              that the Engine was passed, and is called with `*call_args` and `**call_kwargs`
              as parameters. `call` can be omitted altogether, in which case the
              motor's `.goto()` function is assumed.

            name: if not `None`, then the full state will be `BUSY.{name}`. Otherwise
              it's just `BUSY` if `call` is a true callable, or `BUSY.{call}` if `call`
              was passed as a string.

            done_flag: if not `None`, the presence of this flag in `motor.flags` indicates
              that the work has finished.

            busy_flag: if not `None`, the presence of this flag in `motor.flags` indicates
              that the device is still busy performing this task.

            busy_check: if not `None`, it is expected to be a callable which returns `True`
              while the device is still busy peforming the task.

            *call_args, **call_kwargs: named and unnamed arguments to pass to `call`.

        Returns: the result of the `call` call.
        '''

        # go call
        if hasattr(call, "__call__"):
            # callable is an explicit lambda or similar
            callable = call
            
        elif isinstance(call, str):
            # callable is a string-named member of the motor
            callable = getattr(self.__motor, call)
            if name is None:
                name = call

        else:
            # default parameter 'call' is a number, callable is suppsoed to be 'goto',
            # and we need to place the number in front of the call_args list, in fact.
            callable = self.__motor.goto
            call_args = tuple([call] + list(call_args))
            name = "goto"

        # busy call
        if hasattr(busy_check, "__call__"):
            busy_callable = busy_check
            
        elif done_flag is not None:
            busy_callable = lambda: done_flag not in self.__motor.flags
            
        elif busy_flag is not None:            
            busy_callable = lambda: busy_flag in self.__motor.flags
            
        else:
            busy_callable = self.__motor.moves

        self.__scheduled_go = {
            'name': name,
            'proc': partial(callable, *call_args, **call_kwargs),
            'busy': busy_callable
        }

        logger.debug(f"New go: {self.__scheduled_go}")

        
    def clear_motor(self):
        self.__motor.clear()



class WorkerObject(object):
    '''
    Interface for a worker object to be managed by a WorkerEngine.
    '''
    def work(self, params):
        pass

    def abort(self):
        pass

    def clear(self):
        pass

    def isBusy(self):
        pass


class WorkerEngine(object):
    '''
    This models an EPICS device that "does something." It's the precursor of
    a positioner (e.g. a Motor) in the sense that it has a simple state
    diagram which shows whether the device is currently busy performing a
    task ("BUSY"), or free to accept tasks ("IDLE").

    The complete state diagram is as follows:

      - INIT: starting up, ends up in IDLE

      - IDLE: waiting for tasks, can exit through BUSY or ERROR

      - BUSY: currently performing, can exit through DONE or ERROR

      - DONE: done performing, cleaning up / waiting to go to IDLE

      - ERROR: error, user needs to acknowledge

      - FAIL: irrecoverable error

    The state names are configurable.
    '''
    
    def __init__(self, stateNames=None):

        self.states = stateNames or {
            'INIT': 'INIT',
            'IDLE': 'IDLE',
            'BUSY': 'BUSY',
            'DONE': 'DONE',
            'ERROR': 'ERROR',
            'FAIL': 'FAIL'
        }

        # All should return the 
        self.state_workers = {
            'INIT': self.state_INIT,
            'IDLE': self.state_IDLE,
            'BUSY': self.state_BUSY,
            'DONE': self.state_DONE,
            'ERROR': self.state_ERROR,
            'FAIL': self.state_FAIL
        }

        # Initial work do be done when entering a state -- no return value.
        self.state_entries = {
            'INIT': self.enter_state_INIT,
            'IDLE': self.enter_state_IDLE,
            'BUSY': self.enter_state_BUSY,
            'DONE': self.enter_state_DONE,
            'ERROR': self.enter_state_ERROR,
            'FAIL': self.enter_state_FAIL
        }

        self.__state = self.states["INIT"]
        self.__do_run = True
        

    def enter_state_generic(self):
        pass

    # Ignore INIT for now, jump straight to IDLE
    enter_state_INIT = enter_state_generic
    def state_INIT(self):
        return "IDLE"
    
    # FAIL is easy, it does nothing.
    def enter_state_FAIL(self):
        log.error("Entered FAIL -- tttthat's all, folks!")
    def state_FAIL(self):
        return "FAIL"

    # The rest... just wait.
    enter_state_IDLE = enter_state_generic
    enter_state_BUSY = enter_state_generic
    enter_state_DONE = enter_state_generic
    enter_state_ERROR = enter_state_generic

    async def run(self, period=0.1):
        while self.__do_run:
            tstart = time.time()
            current_state = self.__state
            new_state = self.state_workers[current_state]()
            if new_state != current_state:
                logger.debug("State: %r -> %r" % (current_state, new_state))
                self.__state = new_state
                self.state_entries[new_state]()
            tdiff = time.time()-tstart
            await asyncio.sleep(tdiff)
        



#!/usr/bin/python3

import pytest, os, random, string
import emmi.app
import multiprocessing as mp
mp.set_start_method("spawn", force=True)


@pytest.fixture(scope='session')
def session_prefix():
    p = ''.join(random.choice(string.ascii_lowercase) \
                for i in range(6))
    sp = os.environ.get('EMMI_TEST_SESSION_PREFIX', p)
    print(f'Session IOC prefix: "{sp}"')
    return str(sp)


def make_app(*args, Type=None, **kwargs):
    if Type is None:
        Type = emmi.app.IocApplication
    aobj = Type(*args, **kwargs)
    aobj.setupIoc()
    aobj.runIoc()


@pytest.fixture(scope='class')
def ioc_prefix(session_prefix):
    p = f'{session_prefix}:{"".join(random.choice(string.ascii_lowercase))}:'
    print(f'AppSync ioc_prefix: {p}')
    return p

@pytest.fixture(scope='class')
def emmi_app_ioc(ioc_prefix, emmi_app_class):
    p = mp.Process(target=make_app, kwargs={'prefix': ioc_prefix,
                                            'Type': emmi_app_class})
    p.start()
    yield ioc_prefix
    p.kill()
    p.join()



#!/usr/bin/python

from emmi import app
import pytest, random, string, time, asyncio

from .conftest import make_app

from caproto.sync.client import read as ca_read
from caproto.sync.client import read as ca_write

def test_cfgFromEnv():
    flat = {
        'foo_var': 'foo_var',
        'foo_car': 'foo_car',
        'foo_bar_car': 'foo_bar_car',
        'bar_car': 'bar_car'
    }

    conf = app.cfgFromFlat(flat, prefix="foo")
    
    assert len(conf) == 3
    assert len(conf['bar']) == 1
    assert conf['car'] == 'foo_car'
    assert conf['var'] == 'foo_var'


def test_cfgFromEnv_prefix():
    assert app.cfgFromFlat({"foo_bar_moo": 3}, prefix='foo')['bar']['moo'] == 3

def test_cfgFromEnv_noprefix():
    assert app.cfgFromFlat({"foo_bar_moo": 3})['foo']['bar']['moo'] == 3
    


def test_cfgUnify_simple():
    m1 = { 'foo': 1 }
    m2 = { 'bar': 2 }
    m3 = { 'moo': 3 }
    m4 = { 'foo': 4 }
    
    u1 = app.cfgUnify(m1, m2, m3)
    assert u1.get('foo') == 1
    assert u1.get('bar') == 2
    assert u1.get('moo') == 3

    u2 = app.cfgUnify(m1, m2, m3, m4)
    assert u2.get('foo') == 4
    assert u2.get('bar') == 2
    assert u2.get('moo') == 3


def test_cfgUnify_nested():
    m1 = { 'foo': { 'bar': 1 } }
    m2 = { 'foo': { 'moo': 2 } }
    m3 = { 'bar': { 'moo': 3 } }
    m4 = { 'foo': { 'bar': 4 } }
    m5 = { 'foo': 5 }

    u1 = app.cfgUnify(m1, m2, m3)
    assert u1.get('foo').get('bar') == 1
    assert u1.get('foo').get('moo') == 2
    assert u1.get('bar').get('moo') == 3

    u2 = app.cfgUnify(u1, m4)
    assert u1.get('foo').get('bar') == 1
    assert u2.get('foo').get('bar') == 4
    assert u2.get('foo').get('moo') == 2
    assert u2.get('bar').get('moo') == 3

    u3 = app.cfgUnify(u1, m5)
    assert u3.get('foo') == 5
    assert u3.get('bar').get('moo') == 3


def test_cfgUnify_case():
    m1 = { 'fooBar': 1 }
    m2 = { 'FOOBAR': 2 }

    # all-caps FOOBAR will overwrite camel-case fooBar
    u1 = app.cfgUnify(m1, m2)
    assert u1.get('fooBar') == 2

    # all-caps FOOBAR is first, so will produce 'foobar'
    # then camel-case fooBar will be a new entry
    u2 = app.cfgUnify(m2, m1)
    assert u2.get('fooBar') == 1
    assert u2.get('foobar') == 2

    # no case mangling, all keys are kept verbatim
    u3 = app.cfgUnify(m1, m2, mangle_case=False)
    assert u3.get('fooBar') == 1
    assert u3.get('FOOBAR') == 2
    assert u3 == app.cfgUnify(m2, m1, mangle_case=False)

    # This should raise an exception because now we have keys
    # FooBar and fooBar that differ in camel-case, but would be
    # the same in upper-case, before we try to update data using
    # FOOBAR as a key.
    m3 = { 'FooBar': 3 }
    with pytest.raises(RuntimeError) as e:
        app.cfgUnify(m1, m3, m2)


def test_app_config():
    # Tests the addConfig/addNestedConfig/addFlatConfig parts of IocApplication

    cfg1 = { 'epics': { 'prefix': 'FOO' },
             'extra': { 'param': 'cfg1' } }
    
    obj1 = app.IocApplication(prefix='BAR', cfg=cfg1)

    # cfg object overrides __init__ parameter
    assert obj1.conf.get('epics').get('prefix') == 'FOO'
    
    obj2 = app.IocApplication()
    obj2.addNestedConfig(cfg1)

    assert obj2.conf == obj1.conf
    assert obj2.conf == cfg1
    
    cfg3 = { 'extra': { 'param': 'cfg3' } }

    obj2.addNestedConfig(cfg3)
    assert obj2.conf['extra'] == cfg3['extra']

    obj2.addFlatConfig({'extra/param': 'cfg4'}, separator='/')
    assert obj2.conf['extra']['param'] == 'cfg4'

    obj2.addFlatConfig({'extra/param': 'cfg4'}, separator='/', subsection='extra')
    assert obj2.conf['extra'].get('extra').get('param') == 'cfg4'

    # setupIoc() prefix does *not* override an already existing prefix
    # But we can't test now because we can't actually access the real
    # prefix :-( All we can do is access obj1.conf['epics']['prefix'],
    # and that won't actually confirm that setupIoc behaves correctly.
    #obj1.setupIoc(prefix='MOO')
    #assert obj1.prefix == 'FOO'
        
    
def test_app_testmode():
    obj1 = app.IocApplication(cfg={'epics': { 'prefix': 'FOO' }})
    obj2 = app.IocApplication(cfg={'epics': { 'prefix': 'FOOTEST' }})

    assert not obj1.testModeRequested()
    assert obj2.testModeRequested()


@pytest.fixture
def random_prefix():
    return ''.join(random.choice(string.ascii_letters) for i in range(32))
    

def _test_app_sync_defaultioc(random_prefix):
    # Runs the default IOC (i.e. one where only the heartbeat PV is active)
    a = app.IocApplication(prefix='yadda')
    a.setupIoc(prefix=random_prefix)
    a.startIoc()

    assert a.running

    tmax = 3
    tstart = time.time()
    while a.running:
        tdiff = time.time()-tstart
        if tdiff < tmax:
            print ("Exitting in %d..." % (tmax+1-tdiff,))
            time.sleep(1.0)
        else:
            a.stopIoc()

    assert (not a.running)


@pytest.fixture(scope='class')
def emmi_app_class():
    return app.IocApplication


class TestAppAsync:    

    @pytest.mark.asyncio
    async def test_app_async_defaultioc(self, emmi_app_ioc):
        pref = emmi_app_ioc
        from caproto.sync import client
        tstart = time.time()
        while time.time()-tstart < 3.0:
            print (ca_read(f'{pref}heartbeat').data)
            await asyncio.sleep(0.5)

    @pytest.mark.asyncio
    async def test_app_moo(self, emmi_app_ioc):
        pref = emmi_app_ioc
        tstart = time.time()
        while time.time()-tstart < 3.0:
            print (ca_read(f'{pref}heartbeat').data)
            await asyncio.sleep(0.5)

            
class TestAnotherIocTest:
    
    @pytest.mark.asyncio
    async def test_another_app(self, emmi_app_ioc):
        pref = emmi_app_ioc
        tstart = time.time()
        while time.time()-tstart < 3.0:
            print (ca_read(f'{pref}heartbeat').data)
            await asyncio.sleep(0.5)
    



#!/usr/bin/python3

from caproto.asyncio.client import Context as ClientContext
from caproto import ChannelType

import logging, asyncio
logger = logging.getLogger(__name__)

class PvStringFormatter:
    def __init__(self, fmt, **components):
        ''' Follows the PVs in `components` and formats `fmt` according to component names.
        '''
        self.comp_format = fmt
        self.comp_dict = components.copy()
        self._comp_rev_dict = { v:k for k,v in self.comp_dict.items() }
        self.comp_pvs = None
        self.comp_monitors = None
        self.ctx = None
        self.comp_data = {}


    async def connect(self, ctx=None):
        if ctx is not None:
            self.ctx = ctx
        if self.ctx is None:
            self.ctx = ClientContext()

        self.comp_pvs = {
            k:v for k,v in zip(
                self.comp_dict.keys(),
                await ctx.get_pvs(*[f for f in self.comp_dict.values()])
            )
        }

        tmp = {p[0]:p[1].name for p in self.comp_pvs.items()}
        logger.info(f'String formatter PVs: {tmp}')

        self.comp_monitors = {
            k:v.subscribe() for k,v in self.comp_pvs.items()
        }

        logger.info(f'String element monitors: {self.comp_monitors}')

        for m in self.comp_monitors.values():
            m.add_callback(self._incoming)


    def _incoming(self, pv, data):
        pv_name = pv.pv.name
        pv_data = data.data[0]
        pv_key = self._comp_rev_dict[pv_name]
        
        if data.data_type == ChannelType.STRING:
            pv_data = pv_data.decode()
            
        self.comp_data[pv_key] = pv_data
        print(f'{pv_key}: {pv_data}')


    @property
    def current(self):
        try:
            return self.comp_format.format(**self.comp_data)
        except KeyError as e:
            logger.error(f'Not enough keys for "{self.comp_format}": {e}')
            return None



#!/usr/bin/python3

import pytest
from emmi import scpi

import pyvisa

import parse

@pytest.fixture
def huber_device():
    ''' Returns a test device to use. '''

    # Virtual device
    return { 'resource_manager': 'tests/pytest_dev.yaml@sim',
             'device': 'ASRL4::INSTR' }

    # Real device
    #return { 'resource_manager': '@py',
    #         'dev': 'TCPIP::192.168.136.216::INSTR' }

    
def _test_pharos(huber_device):

    print(f'Connecting to device: {huber_device}')

    dev = scpi.MagicHuber(**huber_device)
        
    node = scpi.PropertyNode(dev, 'OSC_OUT_STATE')
    with pytest.raises(RuntimeError):
        print("getter 2:", node())

#    print("shutter:", dev.query('SHUTTER_STATE'))

#    shut = scpi.PropertyNode(dev, 'SHUTTER_STATE',
#                             getter_fmt='{name:s}',
#                             #unwrap='never',
#                             #cast='SHUTTER_CTRL: {:d}, {:d}, {:d}')#
#
#
#                             unwrap='always',
#                             cast=('SHUTTER_CTRL: {:d}', '{:d}', '{:d}'))
#                             
#                             #unwrap='never',
#                             #cast=lambda s: parse.parse('{ignore}{:d}, {:d}, {:d}', s).fixed )
#    
#    print("shutter:", shut())
#
#    print(shut()[0], shut()[1], shut()[2])    
#
#    #print(shut()[0][0], shut()[1][0], shut()[2][0])



#!/usr/bin/python3

import asyncio, time, logging
import logging
logger = logging.getLogger(__name__)

from os import environ

from emmi.api.exports import ExportObjectFromDict

from softioc import softioc, builder, asyncio_dispatcher

'''
Base modules for easier development of EPICS
'''

def loadFromYaml(path):
    from yaml import load, dump
    try:
        from yaml import CLoader as Loader, CDumper as Dumper
    except ImportError:
        from yaml import Loader, Dumper
        
    if (path is not None) and len(path) > 0:
        return load(open(path, 'r').read(), Loader=Loader)

    return {}


def cfgFromFlat(flatCfg, prefix='', splitChar='_'):
    '''
    Returns a (possibly nested) configuration dictionary from a flat
    list of key-value pairs `flatCfg`. The initial list can for instance
    be a list of environment variables.

    Only entries where the key matches a specific prefix are considered.
    `prefix`, however, may be an empty string, which leads to accepting
    the list as given.
    
    Splits variable names at `splitChar` and adds a new dictionary
    level for each split.
    '''

    if not isinstance(prefix, str):
        return {}

    env = {}

    for k,v in flatCfg.items():
        if not k.startswith(prefix):
            continue
        
        parts = k.split(splitChar)

        current_dict = env
        for p in parts[:-1]:
            current_dict = current_dict.setdefault(p, {})
        current_dict[parts[-1]] = v
        
    if len(prefix):
        return env.get(prefix, {})
    else:
        return env


def cfgUnify(*cfgs, mangle_case=True):
    '''
    Returns a unified configuration object by consecutively applying
    dictionary values from `cfgs`. The individual configuration
    dictionaries in `cfgs` can be recursive.

    If `mangle_case` is enabled (default), then a special handling
    of key string case is perfomed, as follows.

    Quite generally, the preferred format of key names is
    all-lower-case ("likethis") or camel-case ("likeThis"), without
    dashes or underscores, and the typical usage case for this function
    is to put together configuration information from different sources,
    e.g. config files (e.g. YAML hierarchies) and environment variables.
    It is acknowledged that some configuration sources (e.g. environment
    variables) have their typical case customs (e.g. ALL_UPPER_CASE_IN_ENV_VARS).

    For this reason, when encountering all-caps keys, corresponding
    existing keys are searched in previous configuration dictionaries.
    If those are found, then the new ALL-CAPS key is interpreted to
    update one of the existing variables. For instance an environment
    variable `APPLICATIONNAME` will overwrite a previous config setting
    `applicationName`.

    If no case-sensitive keys with the same name exist, then new keys are
    created from scratch. But if the newly to-be-created key is all-caps,
    then instead of using it verbatim, it is converted to lower-case (!),
    to keep the style of lower-case variable names consistend.

    Note that while this gives well-defined results, it may not behave as
    expected if configuration keys use camel-case and are expected; for
    instance, "FOOBAR" would be translated to "foobar" if "fooBar"
    doesn't already exist.
    
    Defining camel-case variables in environment variables works, though:
    "$fooBar=application" would be translated to a configuration variable
    "fooBar='application' regardless of where it is found.

    The configuration dictionaries in `cfgs` are applied subsequently,
    meaning the last one wins (i.e. overwrites previously existing values
    for the same key).
    '''

    #print ("Unify:", *cfgs)
    
    cfg = dict()
    for c in cfgs:
        if c is None:
            continue

        # Map of:  CAMELCASEKEY -> camelCaseKey  for the existing dict ('cfg')
        if mangle_case:
            keysCaseMap = { k.upper(): k for k in cfg.keys() }
            if len(keysCaseMap) != len(cfg):
                raise RuntimeError("Oops. You probably have several camel-case configKey items in your "
                                   "config settings that each map to the same all-caps CONFIGKEY. "
                                   "You can't to that.")
            
        for input_k,v in c.items():

            if mangle_case and input_k.isupper():
                # case-mangling: if we have an all-caps input key, we need
                # to find a corresponding cammel-case key to it.
                try:
                    k = keysCaseMap[input_k]
                except KeyError:
                    k = input_k.lower()
                    #print(f'{input_k} is a new one, using it as {k}')
            else:
                # no key-case-mangling
                k = input_k
                
            cfg[k] = v if not isinstance(v, dict) \
                else cfgUnify(cfg.get(k, None), v)
            
    return cfg


class IocApplication(object):
    '''
    Specialized application model for EPICS IOCs.
    
    While frameworks like `pythonSoftIOC` bring a lot of initialization
    code to the table, generally writing a robust application with useful
    features for production concerns itself with more detals, as for
    instance:
    
      - Configuration parameters -- EPICS prefix, timing and polling
        defaults etc.

      - Additional flow variables -- heartbeat and kill switch PVs

      - Keeping track of PV objects (e.g. as created by pythonSoftIOC's
        `builder.aIn/aOut` methods).

      - Ability to rapidly invoke features offered by `emmi.api.exports`,
        e.g. add PVs for the properties of a custom object, possibly based
        on an declarative configuration, e.g. as with
        `emmi.api.exports.ExportObjectFromDict`).

    In creating a "smooth application experience", all these different
    roles have subtle interactions, which are none the less quite generic,
    universal and repetitive for essentially any Python-based IOC.

    This object offers a convenient interface of juggling these requirements.
    
    One of the main premises of the application, beyond the obvious fact that
    it's an IOC, is that every configuration aspect may come from more several
    sources with differing priorites. For instance:
    
       - Default settings with lowest priorites inside the Python code itself

       - On-disk configuration files (e.g. in JSON or YAML format)

       - Environment variables with a specific naming scheme (e.g. prefixed
         by the name of the application), overriding any config file settings

       - Command line arguments, being used with highest priority and overriding
         even enviroment variable settings.

    For this, `IocApplication` holds an internal nested configuration dictionary,
    accessible in the `.conf` attribute. It is of course possible to update the
    dictionary directly. But also the helper functions `addFlatConfig()` and
    `addNestedConfig()` are intended to merge new values form a higher-priority
    dictionary, overriding existing settings where necessary.

    The `.conf` attribute is meant as well for internal use by `IocApplication`
    as for external use and reference by objects that aren't managed by a
    specific `IocApplication` instance.

    Here is a non-exhaustive list of `.conf` variables that `IocApplication`
    reacts to (using the hierarchical sytax for `option.suboption` to reflect
    to entries within the nested dicitonary `.conf["option"]["subobtion"]`):

      - `epics.prefix`: The EPICS PV prefix to use. Ultimately this is used
        by `setupIoc()`, but the latter may be triggered from `__init__()`.

      - `epics.heartbeat`: The interval at which the "::heartbeat" PV is
        updated (see `setupIoc()` for details)

      - `epics.killSwitch`: Whether or not to export a kill switch PV which
        has the ability to shut down the entire IOC remotely (note that there
        may be security implications with this).

      - `epics.defaults`: Default settings for various Connectors (from
        `emmi.api.exports`) that will be passed on from within the
        `exportObject()`  method. See `emmi.api.exports.ExportObjectFromDict()`
        for details, that's what `exportObject()` uses under the hood.

    Configuration settings that aren't used by `IocApplication` are entirely
    the IOC programmer's responsibility. We suggest, however, that one or more
    sub-sections are created within `.conf` to host such settings (e.g.
    `.conf["harp"]` as used by `miniquant-ioc`, an EMMI-based IOC for the
    HydraHarp 400 photon counter by PicoQuant; or `.conf["scpi"]', as used
    by `escpi` a generic SCPI-to-EPICS brige that is built using EMMI and
    `IocApplication`.)

    A minimal `IocApplication` which exports properties of `MyDeviceObject`
    might look like this:

    ```
    #!/usr/bin/python3
    
    from emmi.app import IocApplication
    from emmi.scpi import MagicScpi
    from os import environ
    from json import load as jsn_load

    # very simple example for an object containing a property/attribute
    # that we want to export access to via EPICS PVs in a python IOC
    class MyDeviceObject:
        theProperty = 3.14

    dev = MyDeviceObject()

    # This one might for example contain something like:
    #   { 'epics': { 'prefix': 'BAR' } }
    cfg = jsn_load(open("./simple-ioc.json"))

    # Initialize with default EPICS prefix of FOO.
    # Also suppress the initialisation of the IOC for now, because
    # we may want to load even more configuration options.
    app = IocApplication (prefix="FOO", setupIoc=False)

    # We want to use the settings from JSON file. These will override
    # existing settings with the same key; e.g. after this, the
    # EPICS prefix might actually be set on track for "BAR".
    app.addNestedConfig(cfg)

    # Add support for magic configuration by environment variables.
    # For instance, after this, the variable MYIOC_EPICS_PREFIX=MOO
    # will actually override the PV prefix defined at runtime from "FOO",
    # or "BAR" defined in the config file, to "MOO" defined in the env-var.
    app.addFlatConfig(os.environ, prefix='MYIOC')

    # Initialization of IOC. This will already create a 'FOO::heartbeat'
    # and 'FOO::killSwitch` PV.
    app.setupIoc(killSwitch=True)

    # Add our object-specific PVs -- in this case, for the
    # MyDeviceObject.theProperty attribute.
    app.exportObject(dev, settings={ 'recordType': 'property',
                                     'property': { 'name': 'theProperty',
                                                   'kind': 'analog' } })

    # Finally run the IOC. After this, the PVs are available on
    # the network to be used and abused with caget/caput.
    app.runIoc()
    ```

    This application is actually available verbatim form the `examples/`
    subdirectory of the main `emmi` sources tree -- you are encouraged
    to try this out!

    `IocApplication` works hand in hand with `emmi.api.exports` to make
    it easy to use a declarative interface for which properties/attributes
    to be exported as PVs.
    '''
    
    def __init__(self, prefix=None, cfg=None, setupIoc=False,
                 name=None, env=None, args=None):
        '''
        Initialize an IOC application based on pythonSoftIOC.
        
        Args:
        
            prefix: the variable / device prefix to use for the EPICS device
        
            cfg: a (possibly nested) configuartion dictionary. Currently, the
              following keys are used:
              - `epics.prefix`: the IOC variable / device prefix. If present, this
                overrides the `prefix` parameter from `__init__()`.
              An alternative way of configuring the application is by
              passing over `args` and `env` information.

            setupIoc: if set to `True`, this will automatically call the
              `setupIoc()` method with default parameters, which needs to be done
              before the IOC is ready to accept `exportObject()` and execute.
              However, generally you don't want to do that if you have multiple
              configuration sources (e.g. YAML files, environment variables etc)
              to be loaded in additio to `cfg`, and which may not yet be available
              at the point at which your `IocApplication` is being defined.

            args: Command line arguments for the application, typically from
              `sys.argv`. These are not processed right away, since an argument
              parser is needed for that. They are stored for later processing
              through `.configure()`.
              There are a number of command line arguments which, if present
              in the parser, have a very specific meaning for `IocApplication`.
              See the documentation of `.configure()` for a documentation.
        '''
        
        self.conf = cfg or dict({})

        self._args = args
        self._env = env
        self._name = name or 'emmi'
        #self._logger = logger

        self.otel_logger_provider = self.setup_otel_logger(logger)

        if prefix is not None:
            self._set_prefix(prefix)

        if setupIoc:
            self.setupIoc()


    def setup_otel_logger(self, logger):

        env = self._env
        service = env.get("EMMI_OTELCOL_SERVICE", self._name)
        otelcol_url = env.get("EMMI_OTELCOL_URL", "") # http://localhost:4317

        if (otelcol_url is None) or (len(otelcol_url)==0):
            return None
            
        from opentelemetry._logs import set_logger_provider
        from opentelemetry.exporter.otlp.proto.grpc._log_exporter \
            import OTLPLogExporter
        from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
        from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
        from opentelemetry.sdk.resources import Resource

        otel_logger_provider = LoggerProvider(
            resource=Resource.create(
                {
                    "service.name": service,
                    "service.class": "epics-ioc",
                    "service.instance.id": os.uname().nodename,
                }
            ),
        )
        set_logger_provider(otel_logger_provider)
        otel_otlp_exporter = OTLPLogExporter(endpoint=otelcol_url,
                                             insecure=True)
        otel_logger_provider.add_log_record_processor(
            BatchLogRecordProcessor(otel_otlp_exporter)
        )

        #  console_exporter = ConsoleLogExporter()
        #  logger_provider.add_log_record_processor(BatchLogRecordProcessor(console_exporter))

        otel_handler = LoggingHandler(level=logging.NOTSET,
                                      logger_provider=otel_logger_provider)
        logger.addHandler(otel_handler)

        self.have_otel = True

        return otel_logger_provider


    def configure(self, parser,
                  args=None, args_prefix_map=None,
                  env=None, env_prefix_map=None, env_prefix=None):
        '''
        Reads configuration settings for the application.

        Configuration precedence is enforced through `cfgUnify()`
        and in the following order:

          - The least important setting is one read from a (possibly
            nested) config file, specified through a possible
            `--from-yaml` argument (see the "Args" section below).
        
          - Next in line are environment variables: variables of the
            form `{env_prefix}_{config_keys}={value}` override pre-existing
            keys `{config_keys}` (same caveats with underscore apply),
            or are freshly defined if they didn't exist previously.
        
          - The highest precedence is given to command line arguments,
            specified verbatim. The `cfg` object keys correspond to the
            line arguments, with every underscore opening a `cfg`
            subsection (e.g. `--epics-prefix` sets the
            `cfg["epics"]["prefix"]` key).
        

        Args:
            parser: a configured Python command line parser
        
            args: an arguments array (e.g. `sys.argv`)

            args_prefix_map: if this is not `None`, it acts as an additional
              filter on the command line arguments which we want to consider.
              The general format of a command line argument is
              `--{prefix}-{confkey}`, where `{confkey}` is the (possibly nested,
              by the use of the `-` character) configuration key. The keys in
              `args_prefix_map` are expected to describe the valid `{prefix}`
              values which are considered. Their value counterpart is the
              subtree within the config dictionary where to insert the values.
              Note that the subsection cannot be nested (i.e. it needs to be
              a direct subsection of the config tree). To disable config
              command line arguments either don't pass an `args` object,
              or pass an empty `dict()` here.

            env: environment variables list. If not specified, the one
              passed over to `.__init__()` will be used.

            env_prefix_map: this is used for filtering environment variables in
              a similar fashion to how `args_prefix_map` is used for filtering
              arguments, with one notable difference: if this is `None`,
              then environment variable configurations is disabled completely.
              This is because typically, there are more configuration variables
              that _don't_ concern us than there are those that do.
              So this _must_ contain at least one entry, or env-var config is
              disabled.

            env_prefix: convenience use that creates
              `env_prefix_map = { "...": None }` (i.e. an entry which uses
              the prefix "..." to insert data into the root of the config tree),
              if `env_prefix_map` is `None`.

        Returns: the command-line arguments namespace, for possible subsequent
        use.
        '''
        if args is None:
            args = self._args

        if env is None:
            env = self._env

        pargs = parser.parse_args()

        ## Note that some stuff here should be logged, but generally,
        ## at this point there's no logger available if we want to let
        ## OpenTelemetry handle the logging (because we initialize
        ## OpenTelemetry only after the config has been handled).

        # Least important: config file --from-yaml
        if hasattr(pargs, "from_yaml"):
            #print(f"config={pargs.from_yaml} type=file")
            ycfg = loadFromYaml(pargs.from_yaml)
            self.addNestedConfig(ycfg)

        # Next in line: env-vars with {env_prefix}
        if env_prefix_map is None and env_prefix is not None:
            env_prefix_map = { env_prefix: None }
        if env_prefix_map is not None:
            for k,v in env_prefix_map.items():
                print(f'config={k} type=env subsection={v}')
                self.addFlatConfig(env, prefix=k, subsection=v)

        # Highest priority: direct command line arguments
        if args_prefix_map is None:
            args_prefix_map = { "": None }
        for k,v in args_prefix_map.items():
            #print(f'config="{k}" type=args subsection="{v}"')
            filtered = {k:v for k,v in vars(pargs).items() if v is not None}
            self.addFlatConfig(filtered, prefix=k, subsection=v)

        if (self.otel_logger_provider is None):
            logging.basicConfig(level=self.conf.get('log', {}).get('level', "INFO"))

        return pargs
        

    @property
    def prefix(self):
        return self.conf['epics']['prefix']


    @property
    def epicsPrefix(self):
        logger.warning(f'epicsPrefix is deprecated, use .prefix instead')
        return self.conf['epics']['prefix']


    def _set_prefix(self, prefix):
        # pythonSoftIOC will add a ':' by its own. Meanwhile, we're trying
        # to use prefix _exactly_ as specified (for later compatibility with
        # caproto, for instance).
        if prefix[-1] == ':':
            self.conf.setdefault('epics', {}).setdefault('prefix', prefix)
        else:
            logger.warning(f'EPICS prefix doesn\t have trainling colon ({prefix})'
                           f' -- expecting EMMI to add it is deprecated and will be remove')
            self.conf.setdefault('epics', {}).setdefault('prefix', f'{prefix}:')


    def addNestedConfig(self, cfg):
        self.conf = cfgUnify(self.conf, cfg)


    def addFlatConfig(self, cfg, prefix='', separator='_', subsection=None):
        nested = cfgFromFlat(cfg, prefix, separator)
        self.addNestedConfig(nested if subsection is None else {subsection: nested})


    #def testModeRequested(self):
    #    '''
    #    Returns True if test mode was requested of the IocApplication.
    #    The application itself does not behave differently, but external
    #    modules (e.g. hardware initialisation) might.
    #
    #    NOTE: This should be obsolete pretty soon. It's not a good idea
    #    for code to differentiate its behavior depending on whether
    #    it's supposed to be in test mode or not.
    #    '''
    #    return self.conf['epics']['prefix'].endswith('TEST')
        

    def setupIoc(self, prefix=None, heartbeat=1.0, killSwitch=True):
        '''
        Initializes the IOC part of the application. If `prefix` is specified,
        it overrides whatever prefix was set in the configuration system.
        Note that only one PV is initialized here -- the obligatory 'heartbeat'
        PV that every EMMI IocApplication has and which just implements an
        ever-incrementing counter with a fixed frequency (default: 1 second).
        '''

        # If prefix is not 'None', override the cfg.epics.prefix.
        if prefix is not None:
            self._set_prefix(prefix)

        try:
            p = self.prefix
        except KeyError as e:
            logger.error(f'key="{str(e)}" desc="Did you configure the EPICS prefix?"')
            raise

        # Set some defaults if they're not specified in cfg.epics
        self.conf.setdefault('epics', {}).setdefault('heartbeat', heartbeat)
        self.conf.setdefault('epics', {}).setdefault('killSwitch', killSwitch)


        # We're storing the prefix CAproto style, as-is, including the
        # trailing ':'. But pythonSoftIOC expects it without, and will
        # add the trailing ':' by itself.
        if self.prefix[-1] != ':':
            logger.warning(f'Your prefix doesn\'t contain a trailing colon (:) '
                           f'-- that\'s good, but pythonSoftIOC will add one and there\'s nothing'
                           f' we can do about it for now.')
            pysoftioc_prefix = self.prefix
        else:
            pysoftioc_prefix = self.prefix[:-1:]
            
        logger.info(f'Using EPICS device prefix: "{pysoftioc_prefix}" (that\'s what we\'ll send to pysoftioc)')
        
        self.softioc = softioc
        self.iocBuilder = builder
        self.iocDispatch = asyncio_dispatcher.AsyncioDispatcher()
        self.iocBuilder.SetDeviceName(pysoftioc_prefix)

        # the device property part
        self.pv = {}

        if self.conf['epics']['heartbeat'] is not None:
            self.pv['heartbeat'] = self.iocBuilder.aIn('heartbeat', initial_value=0)

        if self.conf['epics']['killSwitch']:
            self.pv['killSwitch'] = self.iocBuilder.boolOut('killSwitch', initial_value=False,
                                                            on_update=self.die)


    def die(self, really=True):
        '''
        Turns off the runLoop variable if `really` is `True`. This is needed
        for the killSwitch option of the application (which, once activated,
        should be impossible to stop).
        '''
        if really:
            self.runLoop = False


    async def heartbeat(self):
        '''
        This is the "heartbeat" loop of the application.
        '''
        self.runLoop = True
        counter = 0
        while self.runLoop:
            self.pv['heartbeat'].set(counter)
            counter += 1
            await asyncio.sleep(self.conf['epics']['heartbeat'])
        self.running = False


    def startIoc(self, setup=False):
        '''
        Starts the IOC. In the pythonSoftIOC implementation, this just registers
        our own `heartbeat()` loop with the dispatcher and initialises the
        IOC (see `pythonSoftIOC.iocInit()`).
        '''
        if setup:
            self.setupIoc()
        self.iocBuilder.LoadDatabase()
        self.iocDispatch(self.heartbeat)

        logger.info(f'EMMI IOC appliction with EPICS prefix: "{self.prefix}" (EMMI internal representation)')
        self.softioc.iocInit(self.iocDispatch)
        self.running = True


    def stopIoc(self):
        '''
        Signalizes the heartbeat loop that it has to stop. This in turn will make
        the main loop of `runIoc()` stop.
        '''
        self.runLoop = False
    

    def runIoc(self, exit=True, setup=False):
        '''
        Calls startIoc() and waits for it to return (which may be never).
        If `exit` is set to `True` (the default), the pythonSoftIOC `exit()`
        method is called, which kills the application.
        '''
        
        self.startIoc(setup=setup)
        
        while self.running:
            time.sleep(0.1)

        if self.otel_logger_provider is not None:
            self.otel_logger_provider.shutdown()
        
        if exit:
            self.softioc.exit()


    def exportObject(self, device_obj, settings):
        '''
        Takes properties of object `device_object` and generates EPICS PVs
        according to the dicitonary `settings`. `settings` is a (usually YAML-generated)
        declarative format as defined in `emmi.api.exports`.
        This is actually just a wrapper for `emmi.api.exports.ExportObjectFromDict()`.
        '''
        defaults = self.conf.get('epics', {}).get('defaults', None)
        return ExportObjectFromDict(self.iocDispatch, device_obj, settings,
                                    recordTypeDefaults=defaults)



#!/usr/bin/python3

import pytest
from emmi import scpi

import pyvisa

import asyncio

@pytest.fixture
def visa_scpi_device():
    ''' Returns a test device to use. '''

    # Virtual device
    return { 'resource_manager': 'tests/pytest_dev.yaml@sim',
             'device': 'ASRL3::INSTR' }

    # Real device
    return { 'resource_manager': '@py',
             'device': 'TCPIP::192.168.136.216::INSTR' }

    
def test_property_node_get(visa_scpi_device):
    # Testing the PropertyNode object. Essentially, we want to know that
    # naming is correct, get() / set() work as expected with all associated
    # parsing, etc.
    
    rm = pyvisa.ResourceManager(visa_scpi_device['resource_manager'])
    dev = rm.open_resource(visa_scpi_device['device'],
                           read_termination='\n',
                           write_termination='\n' )

    #print("Available devices:", rm.list_resources())
    #print("Dev ID:", dev.query("*IDN?"))

    # single number parameter, current value is "100"
    n1 = scpi.PropertyNode(dev, "DISP:BACK")
    assert n1.name == "DISP:BACK"
    
    # array (2 elements), current value is "-0.0004, -0.0004"
    n2 = scpi.PropertyNode(dev, "FUNC:GVAL")
    assert n2.name == "FUNC:GVAL"

    # default behavior: automatic, "conservative" casting
    assert type(n1()) == int
    assert n1() == 100
    assert len(n2()) == 2
    assert type(n2()[0]) == float

    # unwrapping: single, never, always
    print("N1:", n1(unwrap="always"))
    assert type(n1(unwrap="always")) == tuple
    assert len(n1(unwrap="always")) == 1
    assert n1(unwrap="single") == 100
    with pytest.raises(ValueError):
        n2(unwrap="single")
    assert type(n2(unwrap="never")) == str

    # casting: int, float, callable
    assert type(n1(cast=None)) == str
    assert type(n1(cast="none")) == str
    assert type(n1(cast="auto")) == int
    with pytest.raises(ValueError):
        # cannot parse "100" as float
        assert type(n1(cast="{:f}")) == float
    assert n1(cast=lambda v: "large" if int(v) > 50 else "small") == "large"

    # tuple-casting
    t = n2(unwrap="auto", cast=(float, str))
    assert len(t) == 2
    assert type(t[0]) == float
    assert type(t[1]) == str

    # complex parsing
    cm = n2(unwrap="never", cast="{first}, {second:f}")
    assert type(cm['first']) == str
    assert type(cm['second']) == float


@pytest.mark.asyncio
async def test_async_query(visa_scpi_device):
    rm = pyvisa.ResourceManager(visa_scpi_device['resource_manager'])
    dev = rm.open_resource(visa_scpi_device['device'],
                           read_termination='\n',
                           write_termination='\n' )

    n1 = scpi.PropertyNode(dev, "DISP:BACK") # single number (int, mostly)
    n2 = scpi.PropertyNode(dev, "FUNC:GVAL") # two floats (formatter is {})

    print(f'Sync read: {n1.get()}, {n2.get()}')

    #v1 = await n1.async_get()
    #v2 = await n2.async_get()

    #del n1
    #del n2
    #print("deleted")

    #print(f'Async read: {v1}, {v2}')
    
    

def test_property_node_set(visa_scpi_device):
    # Testing the PropertyNode object.
    
    rm = pyvisa.ResourceManager(visa_scpi_device['resource_manager'])
    dev = rm.open_resource(visa_scpi_device['device'],
                           read_termination='\n',
                           write_termination='\n' )

    n1 = scpi.PropertyNode(dev, "DISP:BACK") # single number (int, mostly)
    n2 = scpi.PropertyNode(dev, "FUNC:GVAL") # two floats (formatter is {})

    n1(37)
    assert n1() == 37

    n2(3.14e-4, 7.62e-4)
    assert len(n2()) == 2
    assert n2()[0] < 3.15e-4 and n2()[0] > 3.13e-4

    #n2(

    
def test_scpi_node_branch(visa_scpi_device):
    # Testing branching off

    rm = pyvisa.ResourceManager(visa_scpi_device['resource_manager'])
    dev = rm.open_resource(visa_scpi_device['device'],
                           read_termination='\n',
                           write_termination='\n' )

    # test device has property DISP:BACK.
    # we start off in DISP, and try to branch off into BACK.
    
    n1 = scpi.PropertyNode(dev, "DISP")

    assert isinstance(n1.BACK, scpi.PropertyNode)

    assert type(n1.BACK()) == int

    
def test_scpi_open(visa_scpi_device):

    # initialize
    device = scpi.MagicScpi(**visa_scpi_device)
    assert isinstance(device.idn(), str)

    # testing first-level subnode
    assert isinstance(device.DISP, scpi.PropertyNode)
    assert isinstance(device.DISP.BACK, scpi.PropertyNode)

    #assert type(device.DISP.BACK()) == int

    assert device.DISP.BACK.name == "DISP:BACK"

    assert type(device.DISP.BACK()) == int
    
    # test error
    with pytest.raises(KeyError) as e:
        getattr(*(scpi.MagicScpi()))
        assert str(e.value) == 'MAGICSCPI_ADDRESS'


#def test_mock_MockPropertyBranch():
#    
#    foo = scpi.MockPropertyBranch()
#    
#    assert type(foo()) is float
#  
#    foo(2.72)
#    assert foo() == 2.72
#    
#    
#def test_MockScpi(): 
#
#    mock = scpi.MockScpi()
#    assert mock.id() == "Mock SCPI Device"



#!/usr/bin/python3

import asyncio, time, logging

from caproto.sync import client as ca_client
from caproto.asyncio.client import Context
from caproto import CaprotoTimeoutError
import numpy as np

import copy

from xarray import DataArray

logger = logging.getLogger(__name__)

class PvRetry(RuntimeError):
    '''
    Raised by GuidedPvReader when the PVs are not yet ready / guide does
    not yet signal readiness for signal readout.
    '''
    def __init__(self, *p, **kw):
        super().__init__(*p, **kw)


class GuidedPvReader:
    '''
    Observes a "guide" variable to determine when a specific EPICS PV signal is
    available, then collects the PV signal (which can come in a list of other PVs).
    '''

    def __init__(self, pv=None, guides=None, prefix='', auto_xarray=True):
        '''Initialises the reader.

        Args:
        
            pv: A single PV, or a list of PVs, to read out. If not
              specified here, it can be specified later.
        
            guides: A dicitonary of guide variable(s) and their respective
              value to use. The `pv` values will be acquired on the first occasion
              when *all* of the guides' values have changed *to* the value specified
              in the dictionary. If the dictionary value is a callable, it will be
              called with the current (i.e. new) guide values as its sole
              parameters and the `pv` value will be obtained the first time the
              return value changes to `True`.

            prefix: If specified, it will be prepended to all of the
              PVs' and guides' EPICS names.
        '''
        
        self.prefix = prefix or ''
            

        # PVs get a special treatment (e.g. extending _SIGNAL to also acquire the
        # corresponding _OFFSET and _DELTA PVs, if available)
        self.pv = [pv,] if isinstance(pv, str) else [i for i in (pv or []) ]
        if auto_xarray:
            for p in self.pv:
                if p.endswith("_SIGNAL"):
                    self.pv.append(p.replace("_SIGNAL", "_OFFSET"))
                    self.pv.append(p.replace("_SIGNAL", "_DELTA"))
        self.pv = tuple(self.pv)
    
        
        self.guides = {} if guides is None \
            else { prefix+k: v if hasattr(v, "__call__") else lambda x: x == v \
                   for k,v in guides.items() }

        # map EPICS name -> current True/False  evaluation of the guide signal.
        # Note that this is _not_ the guide trigger evaluation, i.e. the condition
        # of whether the waiting for this guide is finished and we're ready to
        # return data! For the latter to be fulfilled, the corresponding
        # guide_eval needs to be changing from 'False' to 'True'!
        self.guide_evals = { k:None for k in self.guides }

        
    def extract_data(self, response, pvName=None, others=None):
        '''
        Extracts "useful" data out of a response telegram.
        '''

        if response is None:
            return None

        if others is None:
            others = {}

        # Channel types can be: CHAR, DOUBLE, FLOAT, STRING, ENUM, LONG, INT.
        # The intention is to get an automatic useful native Python data type,
        # scalar or array. This means different things for different data
        # types.
        # In addition, we implement some heuristics to decorate waveforms
        # (== arrays) if our obscure array markers are present (shape, dimensions,
        # axis scaling -- to be documented ;-) )
        
        if response.data_type in (ca_client.ChannelType.STRING,):
            return response.data[0].decode('utf-8')
        
        elif response.data_type in (ca_client.ChannelType.DOUBLE,
                                    ca_client.ChannelType.FLOAT,
                                    ca_client.ChannelType.LONG,
                                    ca_client.ChannelType.INT,
                                    ca_client.ChannelType.ENUM):
            
            if len(response.data) == 1:
                return response.data[0]

            if not pvName or not pvName.endswith("_SIGNAL"):
                return response.data
            
            # If we have an array and it ends on _SIGNAL, we also try to
            # load _OFFSET and _DELTA for intrinsic scaling information
            o_name = pvName.replace("_SIGNAL", "_OFFSET")
            d_name = pvName.replace("_SIGNAL", "_DELTA")

            if o_name in others:
                offs = self.extract_data(others.get(o_name, 0))
                #del others[o_name]
            else:
                offs = 0

            if d_name in others:
                dlta = self.extract_data(others.get(d_name, 1))
                #del others[d_name]
            else:
                dlta = 1
                
            axis = offs+np.array(range(len(response.data)))*dlta
                
            return DataArray(data=response.data, dims=["x"], coords=[axis])

        # else: how to handle ENUM / CHAR?
            
        else:
            logger.warning ("Unhandled data type: %r" % (response.data_type,))
            return response.data[0]

    
    def retr(self, pv=None, raiseRetry=True):
        ''' Synchronously checks the guides for readiness and retrieves the PV values.
        
        If `pv` is not `None`, they will be retrieved in addition to the ones
        already specified at the initialisation of the class. If `prefix` is
        specified (not `None`), it will override whatever was specified at the
        initialisation of the class, but only for the PVs specified here.
        '''

        good_guides = 0

        for (k,v) in self.guides.items():
            data = self.extract_data(ca_client.read(k))
            if v(data) and (not self.guide_evals[k]):
                good_guides += 1
            self.guide_evals[k] = v(data)

        if good_guides == len(self.guides):
            pv = [k for k in (pv or {}) ]
            pv.extend([k for k in self.pv])
            return { k: self.extract_data(ca_client.read(self.prefix+k), pvName=self.prefix+k) \
                     for k in pv }

        raise PvRetry()


    async def value(self, timeout=-1, pollPeriod=0.001):
        '''
        Asynchronousluy waits for retr() to deliver a valid dataset.
        Cancels after `timeout` seconds (if timeout >= 0).

        Note that this is _very_ inefficient,, as it essentially
        wraps a (slow) synchronous client in an async loop.
        '''
        tstart = time.time()
        while True:
            try:
                return self.retr()
            except PvRetry:
                if (timeout > 0) and (time.time()-tstart >= timeout):
                    raise
            await asyncio.sleep(pollPeriod)


class GuidedAsyncReader(GuidedPvReader):
    ''' Similar to GuidedPvreader but uses the asyncio caproto client.

    To get as efficient as possible, this implementation actually
    uses subscriptions to the guide variables, and reacts on
    those. Calling .value() will wait asynchronously for the guide
    conditions to be fulflled, and this will be fairly efficient.

    But the most efficient way will be to subscribe a data
    processing callback using .subscribe(), which will be called
    only when all guide conditions are fulfilled and all data
    is available.
    '''
    
    def __init__(self, ctx, *args, subscribe=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.ctx = ctx

        self.data_pvs = None        
        self.guide_pvs = None
        self.subscribe = subscribe
        self.incoming_hooks = []
        self.incoming_lock = asyncio.Lock()


    def subscribe_incoming(self, proc):
        ''' Registers a hook for processing incoming data.
        The hook will receive a dictionary with full PV
        name(s) as keys, and data as values.
        '''
        self.incoming_hooks.append(proc)

        
    async def connect(self, ctx=None):
        
        if ctx is not None:
            self.ctx = ctx
            
        if self.ctx is None:
            logger.info("Initializing CA client context...")
            self.ctx = Context()
            self.own_ctx = True
        else:
            logger.info(f"Using context: {self.ctx}")
        
        guides = [k for k in self.guides.keys()]
        pvnames = [self.prefix+k for k in self.pv ]

        # map EPICS name -> Context PV, for all guide variables
        self.guide_pvs = await self.ctx.get_pvs(*guides)

        # map EPICS name -> Context PV, for all non-guide variables
        self.data_pvs = await self.ctx.get_pvs(*pvnames)
            
        #  map EPICS name -> Subscription obj, for all guides
        self.guide_subscriptions = {}
        self.data_subscriptions = {}

        if self.subscribe:
            for g in self.guide_pvs:
                logger.info(f'Subscribing to guide: {g.name}')
                self.guide_subscriptions[g.name] = g.subscribe()
                self.guide_subscriptions[g.name].add_callback(self._guide_changed)

            for d in self.data_pvs:
                logger.info(f'Subscribing to data: {d.name}')
                self.data_subscriptions[d.name] = d.subscribe()
                self.data_subscriptions[d.name].add_callback(self._data_changed)
        

        # map EPICS name -> trigger-condition-fulfilled, for all guides.
        # When all of these are True, that's when we need to read
        # all the non-guide data.
        # Each of these becomes True when the guide changes value _to_
        # whatever was specified. It switches to False again when a data
        # readout is performed, or when it swithces value away _from_
        # whatever was specified.
        #
        # Note that this is subtly different from self.guide_evals!
        # While guide_evals is a True/False value per guide, representing
        # the current condition evaluation, the guide_trigger truth value
        # also indicates the *changing* to the True value.
        #
        # Also important to note: keeping this accurate for _all_ guides
        # at once is difficult when more than one guide is involved (values
        # will change with a slight delay). Hence the policy: once a guide
        # is considered triggered, it _stays_ triggered for as long as
        # its value doesn't chance, or data isn't read.
        self.guide_trigger = {k:False for k in self.guides.keys()}

        self._incoming_data = {f'{self.prefix}{k}':None for k in self.pv}


    async def disconnect(self):
        if hasattr(self, "own_ctx") and self.own_ctx == True:
            self.ctx.disconnect()
        #for k, v in self.guide_subscriptions.items():
        #    v.unsubscribe()
        #for k, v in self.data_subscriptions.items():
        #    v.unsubscribe()
            

    async def wait_for_guides(self, PVs, timeout=-1):
        ''' Waits for the internal list of PVs '''
        guides = await asyncio.gather(*[v.read() for v in PVs ],
                                      return_exceptions=True)
        for (k,v),data in zip(self.guides.items(),guides):
            if isinstance(data, Exception):
                logger.warning(f'Guide {k}: {data}')
                continue
            self._guide_changed(None,data,pv_name=k)
            if all(self.guide_trigger.values()):
                return len(self.guide_trigger)
        return 0


    def _guide_changed(self, sub, response, pv_name=None):
        ''' Called when a guide value changes. '''

        if pv_name is None:
            pv_name = sub.pv.name

        d = self.extract_data(response)
        eval_result = self.guides[pv_name](d)
        
        #print(f'got {pv_name}: {d}, '
        #      f'eval: {eval_result} <- {self.guide_evals[pv_name]}, '
        #      f'trigger: {self.guide_trigger[pv_name]}')

        if eval_result:
            if not self.guide_evals[pv_name]:
                # on eval switch False -> True: trigger!
                self.guide_trigger[pv_name] = True
        else:
            # eval False always kills the trigger
            self.guide_trigger[pv_name] = False
            
        self.guide_evals[pv_name] = eval_result

        # check whether we can do a data readout (all triggers must be True)
        if not all(self.guide_trigger.values()):
            return

        # If we don't have any hooks, we ignore this part; the data
        # is likely to be retrieved by other means (e.g. by polling
        # in .value()).
        if len(self.incoming_hooks):
            data = self._get_incoming()
            for proc in self.incoming_hooks:
                proc(data)


    def _data_changed(self, sub, response):
        ''' Called when new non-guide data arrives. '''
        self._incoming_data[sub.pv.name] = response

    
    def _get_incoming(self):
        ''' Returns the currently incoming data (from subscriptions)
        and clears everything for the next run.
        '''
        
        for k in self.guide_trigger:
            self.guide_trigger[k] = False

        # need to work on a copy here because _incoming_data might change
        # while we're waiting for new incoming data.
        tmp = self._incoming_data
        #tmp.update(self._incoming_data)
        #print(f'tmp: {len(tmp)}, {[k for k in tmp.keys()]}')

        try:
            data = { k:self.extract_data(v, k, others=tmp) \
                     for k,v in tmp.items() if v is not None }
            return data            
        except RuntimeError as e:
            raise
            #print(f'tmp now: {len(tmp)}, {[k for k in tmp.keys()]}')

        # FIXME: should we clear incoming data, too?...

    
    async def wait_for_incoming(self):
        ''' Waits for incoming data -- this is similar to .value(),
        only this uses the subscription mechanism.
        '''
        await self.wait_for_guides(self.guide_pvs)
        return self._get_incoming()
        

    async def read_data(self):
        ''' Unconditionally reads all (non-guide) data and returns a dictionary.

        Should only be used in a subscription-less mode, after .wait_for_guides().
        '''

        if self.subscribe:
            RuntimeError(f'This is a subscription-based reader; '
                         f'you shouldn\'t use .read_data() here.'
                         f'Use .wait_for_incoming() or '
                         f'.subscribe_incoming() instead.')

        async def read_wrapper(pv):
            try:
                return (pv.name, await pv.read())
            except CaprotoTimeoutError as e:
                logger.info(f'Timeout: {e}')
            
        dataMap = dict(await asyncio.gather(*[read_wrapper(v) for v in self.data_pvs ]))
        
        if not all(self.guide_trigger.values()):
            logger.warning(f'Some triggers were already inactive by the end of readout')
            
        for k in self.guide_trigger:
            self.guide_trigger[k] = False

        return { k:self.extract_data(v, k, others=dataMap) for k,v in dataMap.items() }

    
    async def value(self, timeout=-1, pollPeriod=0.001, autoconnect=True):
        
        if (self.guide_pvs is None) and \
           (self.data_pvs is None) and \
           autoconnect:
            await self.connect()

        if self.subscribe:
            raise RuntimeError(f'This is a subscription-based reader; '
                               f'you shouldn\'t use .value() here.'
                               f'Use .wait_for_incoming() or '
                               f'.subscribe_incoming() instead.')

        tstart = time.time()

        while True:

            # FIXME: should change this to subscriptions!
            if (await self.wait_for_guides(self.guide_pvs)):
                return await self.read_data()

            if (timeout > 0) and (time.time()-tstart >= timeout):
                raise PvRetry()

            await asyncio.sleep(pollPeriod)

            
    def retr(self):
        raise NotImplemented("retr() not available in async mode")


class AsyncMonitorMany:
    ''' Monitors a number of EPICS variables and calls a callback when any one changes.
    '''

    def __init__(self, ctx=None, pvs=None, prefix='', subscribe=True):
        '''Initialises the monitor.

        Args:
        
            pvs: A list of PVs (strings)
      
            prefix: If specified, it will be prepended to all of the
              PVs' and guides' EPICS names.

            subscribe: Monitor using EPICS subscription if True.
        '''
        
        self.prefix = prefix or ''
        self.ctx = ctx
        self.subscribe = subscribe

        # PVs get a special treatment (e.g. extending _SIGNAL to also acquire the
        # corresponding _OFFSET and _DELTA PVs, if available)
        self.pv_names = [pvs,] if isinstance(pvs, str) else [i for i in (pvs or []) ]
        self.pv_names = tuple(self.pv_names)
        
        self.data_pvs = None
        self.incoming_hooks = []
        

    def subscribe_incoming(self, proc):
        ''' Registers a hook for processing incoming data.
        The hook will receive a dictionary with full PV
        name(s) as keys, and data as values.
        '''
        self.incoming_hooks.append(proc)


    async def connect(self, ctx=None):
        
        if ctx is not None:
            self.ctx = ctx
            
        if self.ctx is None:
            logger.info("Initializing CA client context...")
            self.ctx = Context()
        else:
            logger.info(f"Using context: {self.ctx}")

        pvnames = [self.prefix+k for k in self.pv_names ]
        self.data_pvs = await self.ctx.get_pvs(*pvnames)

        self.data_subscriptions = {}

        if self.subscribe:
            for d in self.data_pvs:
                logger.info(f'Subscribing to data: {d.name}')
                self.data_subscriptions[d.name] = d.subscribe()
                self.data_subscriptions[d.name].add_callback(self._data_changed)

        self._incoming_data = {f'{self.prefix}{k}':None for k in self.pv_names}
        

    def _data_changed(self, sub=None, response=None, name=None):
        ''' Called when new data arrives. '''

        pv_name = sub.pv.name if sub is not None else name   
        #logger.debug(f'New data for {pv_name}: {response}')
        
        self._incoming_data[pv_name] = response
        if len(self.incoming_hooks):
            data = self._get_incoming()
            for proc in self.incoming_hooks:
                proc(data)

    
    def _get_incoming(self):
        ''' Returns the currently incoming data (from subscriptions)
        and clears everything for the next run.
        '''
        
        data = { k:v for k,v in self._incoming_data.items() if v is not None }
        return data


    async def wait_for_incoming(self, PVs):
        ''' Waits for incoming data. '''
        try:
            responses = await asyncio.gather(*[v.read() for v in PVs ])
            for pv,res in zip(PVs,responses):
                self._data_changed(sub=None, response=res, name=pv.name)
            
        except Exception as e:
            logger.error(e)
            logger.error(traceback.format_exc())
            self.do_run = False


    async def monitor_loop(self, period=0.01):
        self.do_run = True
        try:
            while self.do_run:
                await self.wait_for_incoming(self.data_pvs)
                await asyncio.sleep(period)
                
        except Exception as e:
            logger.error(e)
            logger.error(traceback.format_exc())
            #self.do_run = False

        logger.error("Oops. Blew it.")



#!/usr/bin/python3

import pytest
from emmi import scpi

import pyvisa

import parse

@pytest.fixture
def pharos_device():
    ''' Returns a test device to use. '''

    # Virtual device
    return { 'resource_manager': 'tests/pytest_dev.yaml@sim',
             'device': 'ASRL1::INSTR' }

    # Real device
    #return { 'resource_manager': '@py',
    #         'dev': 'TCPIP::192.168.136.216::INSTR' }

    
def _test_pharos(pharos_device):

    dev = scpi.MagicPharos(**pharos_device)
    
    #rm = pyvisa.ResourceManager(pharos_device['resource_manager'])
    #dev = rm.open_resource(pharos_device['dev'],
    #                       read_termination='\n',
    #                       write_termination='\r')

    assert "PHAROS UART" in dev.kdev.query('VERSION')

    node = dev.OSC_OUT_STATE

    print("Version:", dev.kdev.query('VERSION'))
    print("Query:", dev.kdev.query('OSC_OUT_STATE'))
    print("Node:", node())

    assert type(node()) == int

    #print ("getter 1:", dev.query('OSC_OUT_STATE'))

    #node = scpi.PropertyNode(dev, 'OSC_OUT_STATE',
    #                         getter_fmt='{name:s}',
    #                         unwrap='always',
    #                         cast='OSC_OUT_CTRL: {:d}')

    #print("getter 2:", node())

    #print("shutter:", dev.query('SHUTTER_STATE'))

    #shut = scpi.PropertyNode(dev, 'SHUTTER_STATE',
    #                         getter_fmt='{name:s}',
    #                         #unwrap='never',
    #                         #cast='SHUTTER_CTRL: {:d}, {:d}, {:d}')
    #
    #                         unwrap='always',
    #                         cast=('SHUTTER_CTRL: {:d}', '{:d}', '{:d}'))
    #                         
    #                         #unwrap='never',
    #                         #cast=lambda s: parse.parse('{ignore}{:d}, {:d}, {:d}', s).fixed )
    # 
    #print("shutter:", shut())
    #
    #print(shut()[0], shut()[1], shut()[2])    
    #
    ##print(shut()[0][0], shut()[1][0], shut()[2][0])



#!/usr/bin/python3

from emmi.app import IocApplication
from emmi.scpi import MagicScpi
from os import environ
from json import load as jsn_load

# very simple example for an object containing a property/attribute
# that we want to export access to via EPICS PVs in a python IOC
class MyDeviceObject:
    theProperty = 3.14

dev = MyDeviceObject()

# This one might for example contain something like:
#   { 'epics': { 'prefix': 'BAR' } }
cfg = jsn_load(open("./simple-ioc.json"))

# Initialize with default EPICS prefix of FOO.
# Also suppress the initialisation of the IOC for now, because
# we may want to load even more configuration options.
app = IocApplication (prefix="FOO", setupIoc=False)

# We want to use the settings from JSON file. These will override
# existing settings with the same key; e.g. after this, the
# EPICS prefix might actually be set on track for "BAR".
app.addNestedConfig(cfg)

# Add support for magic configuration by environment variables.
# For instance, after this, the variable MYIOC_EPICS_PREFIX=MOO
# will actually override the PV prefix defined at runtime from "FOO",
# or "BAR" defined in the config file, to "MOO" defined in the env-var.
app.addFlatConfig(environ, prefix='MYIOC')

# Initialization of IOC. This will already create a 'FOO::heartbeat'
# and 'FOO::killSwitch` PV.
app.setupIoc(killSwitch=True)

# Add our object-specific PVs -- in this case, for the
# MyDeviceObject.theProperty attribute.
app.exportObject(dev, settings={ 'recordType': 'property',
                                 'property': { 'name': 'theProperty',
                                               'kind': 'analog' } })

# Finally run the IOC. After this, the PVs are available on
# the network to be used and abused with caget/caput.
app.runIoc()



#!/usr/bin/python

from schema import *
from softioc import builder as IocBuilder
from softioc import alarm as ioc_alarm

from pprint import pprint
from functools import partial, reduce

import time, asyncio, logging, inspect, sys
logger = logging.getLogger(__name__)

from softioc.softioc import safeEpicsExit as softioc_exit

import numpy as np

import traceback

class EmmiPvDataError(Exception):
    def __init__(self, alrm_severity, alrm_type, msg=None):
        self.alarm_severity = alrm_severity
        self.alarm_type = alrm_type
        super().__init__(msg)

'''
Validators are a concept that allow us to filter, translate and validate
data (values) from one ecosystem (e.g. EPICS) into another (e.g. SCIP).
Each validator needs to have at least:

  - one setup() function which receives named arguments with the necessary
    data for validating; arguments names will be directly accepted/visible
    in the YAML files, so choose wisely.

  - one __getitem__() function which returns the translated value, or
    raises ValueError if the value is out of the permissible conditions.
'''
        
class MappingValidator(object):
    '''
    Translates a value from one set to another
    '''
    def __init__(self, values={}):
        self.setup(values)

    def setup(self, values={}):
        self.trmap = values

    def __getitem__(self, val):
        if val in self.trmap:
            return self.trmap.get(val)
        
        raise ValueError("%r: invalid value" % val)


class SetValidator(object):
    '''
    Accepts only objects form a specific set.
    '''
    def __init__(self, **kwargs):
        self.setup(**kwargs)

    def setup(self, values=[]):
        self.vset = values

    def __getitem__(self, val):
        if val in self.vset:
            return val
        else:
            raise ValueError("%r: value not in set %r" % (val, self.vset))


class BoundaryValidator(object):
    '''
    Accepts only values within specified boundaries
    '''
    def __init__(self, **kwargs):
        self.setup(**kwargs)

    def setup(self, limits=[None, None], inclusive=[True, True], invert=False):
        self.lim = limits
        self.incl = inclusive
        self.inv = invert

    def __getitem__(self, val):
        err = ValueError("%r: outside of range" % val)
        if ((self.lim[0] is not None) and (self.incl[0] and val < self.lim[0])) == (not self.inv):
            raise err
        if ((self.lim[1] is not None) and (self.incl[1] and val > self.lim[1])) == (not self.inv):
            raise err
        if ((self.lim[0] is not None) and (not self.incl[0] and val <= self.lim[0])) == (not self.inv):
            raise err
        if ((self.lim[1] is not None) and (not self.incl[1] and val >= self.lim[1])) == (not self.inv):
            raise err
        return val

    
class SliceValidator(object):
    '''
    Returns only specified slice of an (iterable/indexable) value.
    '''
    def __init__(self, **kwargs):
        self.setup(**kwargs)
        
    def setup(self, start=None, stop=None, step=None):
        self.sobj = slice(start, stop, step)

    def __getitem__(self, val):
        return val[self.sobj]

    
class IdentityValidator(object):
    ''' Lets everything through '''

    def setup(self):
        pass
    
    def __getitem__(self, val):
        return val



class PropertySetter(object):
    '''
    Wrapper that sets a property -- we need this because
    we want to build most of our API based on properties
    (insted of callable setter functions), but at several
    points in the involved APIs (softioc.Builder, for instance)
    callables are required.
    '''

    def __init__(self, name, setter, validator=None):
        self.name = name
        self.setter = setter
        self.validator = validator or IdentityValidator()
        self.validation_failed = False

    async def __call__(self, value):
        try:
            valid = self.validator[value]            
            #logger.info("Value %r validates to %r (by %r)" % (value, valid, self.validator))
            
            self.validation_failed = False
            
            try:
                await self.setter(valid)
                
            except TypeError as t:
                #logger.error(traceback.format_exc())
                self.setter(valid)

            except OverflowError:
                raise
                
        except ValueError as e:
            if not self.validation_failed:
                logger.error(f'desc="Failed to set" property="{self.name}" error="{e}" value="{value}"')
                logger.error(traceback.format_exc())                
                self.validation_failed = True

        except Exception as e:
            logger.error(f'desc="Failed to set" property="{self.name}" error="{e}" value="{value}"')
            logger.error(traceback.format_exc())
                
        except:
            logger.error(f'desc="Failed to set" property="{self.name}" value="{value}"')


class PropertyUpdater(object):
    # FIXME: Better Name: PropertyLoop, because it runs forever, not just once.
    
    '''
    A callable object which loops indefinitely reading out a property via `getter`
    and publishing it on EPICS process-variable `pv`. The waiting
    time on every loop run is such that a full run is as close
    as possible to `period`.
    '''
    
    def __init__(self, name, pv, pollPeriod, getter, validator=None, kill_on_error=False):
        self.name   = name
        self.pv     = pv
        self.period = pollPeriod
        self.getter = getter
        self.validator = validator or IdentityValidator()
        self.kill_on_error = kill_on_error

    def set_period(self, val):
        self.period = val

    async def __call__(self):
        logger.debug("%s: Updater running for PV %r" % (self.name, self.pv))
        available = True
        on_alarm = False
        while True:
            tstart = time.time()
            try:
                val = self.getter()
                
                if (val is not None):
                    cur = self.pv.get()
                    vv = self.validator[val]                    

                    try:
                        is_same = (np.array(cur) == np.array(vv)).all()
                    except Exception as e:
                        logger.error(f'{self.name}: Error comparing current value '
                                     f'{cur} to new value {vv}: {str(e)}, {traceback.format_exc()}')
                        logger.error(f'If this error does not reoccur, you can safely ignore it.')
                        is_same = False

                    if (not is_same) or on_alarm:
                        self.pv.set(vv)

                if not available:
                    logger.info("%s: has become available again" % self.name)
                    available = True

            except EmmiPvDataError as e:
                self.pv.set_alarm(e.alarm_severity, e.alarm_type, time.time())
                on_alarm = True
                
            except Exception as e:
                
                if self.kill_on_error:
                    logger.error(f"{self.name}: property error and IOC termination requested")
                    logger.error(f"{self.name}: {str(e)}")
                    softioc_exit()
                
                # We're not going to exit on PV setting problems (may be
                # intermittent), but at least we'll get loud about it.
                if available:
                    logger.error("%s: property became unavailable, reason: %r" % (self.name, e))
                    logger.error(f"{self.name}: traceback: {traceback.format_exc()}")
                    logger.error("%s: further errors will be silenced" % self.name)
                    logger.info("%s: Getter object was: %r" % (self.name, self.getter,))
                available = False
            
            except e:
                logger.error("%s: unknown exception %r. You _should_ be worried." % e)
                
            finally:
                tdiff = time.time()-tstart
                await asyncio.sleep(self.period - tdiff)

            
'''
List of connector kinds we support. "in"/"out" should return the
correct builder function (lambda parameter being the softioc builder
object itself), and "args" is a list of initial arguments for the
builder function. Typically you'd want to set default values,
and/or sizes, and/or data types.
'''
ConnectorKinds = {
    "analog": {
        "in": IocBuilder.aIn,
        "out": IocBuilder.aOut,
        "create": { "initial_value": 0.0 },
        "to-epics": IdentityValidator(),
        "from-epics": IdentityValidator() },
    
    "switch": {
        "in": IocBuilder.boolIn,
        "out": IocBuilder.boolOut,
        "create": { 'ZNAM': 'OFF', 'ONAM': 'ON' },
        "to-epics": MappingValidator(values={"ON": 1, True: 1, 1: 1,
                                             "OFF": 0, False: 0, 0: 0}),
        "from-epics": MappingValidator(values={1: "ON",
                                               0: "OFF"}) },

    "strings": {
        "in": IocBuilder.stringIn,
        "out": IocBuilder.stringOut,
        "create": { },
        "to-epics": SetValidator,
        "from-epics": SetValidator },

    "waveform": {
        "in": IocBuilder.WaveformIn,
        "out": IocBuilder.WaveformOut,
        "create": { },
        "to-epics": IdentityValidator(),
        "from-epics": IdentityValidator()
    },

    # untested:
    
    "text":  {
        "in": IocBuilder.stringIn,
        "out": IocBuilder.stringOut,
        "create": { },
        "to-epics": SliceValidator(stop=39),
        "from-epics": IdentityValidator() },

    "map": {
        "in": IocBuilder.stringIn,
        "out": IocBuilder.stringOut,
        "create": { "initial_value": "n/a" },
        "to-epics": MappingValidator,
        "from-epics": MappingValidator },

    "values": {
        "in": IocBuilder.longIn,
        "out": IocBuilder.longOut,
        "create": { "initial_value": 0 },
        "to-epics": SetValidator,
        "from-epics": SetValidator },

    "integer": {
        "in": IocBuilder.longIn,
        "out": IocBuilder.longOut,
        "create": { "initial_value": 0 },
        "to-epics": IdentityValidator(),
        "from-epics": IdentityValidator() },
}

class SignalConnector(object):
    '''
    Connects a sensoric signal (i.e. a reading of a measurement) to an
    EPICS PV. The data is retrieved from either a Python property or
    from a getter function.
    '''

    argnames = { "suffix": str,
                 "name": str,
                 "kind": str,
                 "access": object,
                 "create": dict,
                 "validator": dict,
                 "pollPeriod": float,
                 "killOnError": bool }

    schema = Schema({Optional(k):v for k,v in argnames.items()})

    def __init__(self, ioc_dispatch, **kw):
        '''
        Parameters:
          - `ioc_dispatch`: dipatcher to use

          - `name`: Used to recognize stuff throughout the application,
            also serves as default to `suffix`.

          - `suffix`: PV name to use; defaults to `name`.
        
          - `access`: Callable or (obj, propname) tuple of a Python
            property to set or get data from.

          - `kind`: one of the `eda.ConnectorKinds` keys

          - `pollPeriod`: poll period for the data to publish (i.e.
            calling of `access`)

          - `validator`: If not `None`, this is a dict that will be used to
            initialize the validator of the property.

          - `create`: Supplementary arguments to pass on when creating the
            PV class. This is highly specific to the underlying EPICS
            framework being used.
        '''
        
        if not set(kw).issubset(set(self.argnames)):
            raise TypeError("Unknown arguments: %r" % set(kw).difference(self.argnames))
        
        k = ConnectorKinds[kw.get('kind')]

        _s = kw.get('suffix')
        suffix = _s if _s is not None else kw.get('name')

        pv_create_args = k.get('create', {})
        pv_create_args.update(kw.get('create', {}))
        RecordType = k["in"]
        self.pv = RecordType(suffix, **(pv_create_args))

        if hasattr(kw.get('access', None), "__call__"):
            access = kw['access']
        else:
            access = partial(getattr, *(kw['access']))

        # sometimes the validator in ConnectorKinds is an explicit
        # instance (which we can use right away); sometimes it's just
        # a class type, which we have to instantiate first
        valdtr = k['to-epics'](**(kw.get('validator', {}))) if type(k['to-epics']) == type \
            else k['to-epics']

        getter = PropertyUpdater(name=kw.get('name', suffix),
                                 validator=valdtr,
                                 pv=self.pv,
                                 pollPeriod=kw.get('pollPeriod', 1.0),
                                 getter=access,
                                 kill_on_error=kw.get('killOnError', False))

        if ioc_dispatch:
            ioc_dispatch(getter)

        logger.info("Signal: %s, kind='%s', access=%r" % (suffix, kw.get('kind'), kw.get('access')))
        

class ActorConnector(object):
    '''
    Connects an action signal (i.e. button, knob, setting etc) to an EPICS PV.
    '''

    argnames = { "suffix": str,
                 "name": str,
                 "kind": str,
                 "access": object,
                 "create": dict,
                 "validator": object }

    schema = Schema({Optional(k):v for k,v in argnames.items()})

    
    def __init__(self, ioc_dispatch, **kw):
        '''
        Parameters:

          - `ioc_dispatch`: The pythonSoftIOC async dispatcher. Ignored here,
            but we need this for API uniformity with the other connectors.

          - `name`: Name of the ActorConnector (used internally for displaying
            purposes, and as a default for `suffix` if the latter is not specified).
        
          - `suffix`: PV name to use
        
          - `access`: Reference to a Python property (either callable or writable),
            for setting data.

          - `kind`: one of the `eda.ConnectorKinds` keys

          - `validator`: property dictionary for kind-specific validator
            configuration.

          - `create`: Supplementary arguments to pass on when creating the
            PV class. This is highly specific to the underlying EPICS
            framework being used.        
        '''

        if not set(kw).issubset(set(self.argnames)):
            raise TypeError("Unknown arguments: %r" % set(kw).difference(self.argnames))
        
        k = ConnectorKinds[kw.get('kind')]

        _s = kw.get('suffix')        
        suffix = _s if _s is not None else kw.get('name')

        # For debugging
        #def set_attr(*args, **kwargs):
        #    print(f'Set: {args} / {kwargs}')
        #    setattr(*args, **kwargs)
        
        if hasattr(kw['access'], "__call__"):
            access = kw['access']
        else:
            access = partial(setattr, *(kw['access']))
            #access = partial(set_attr, *(kw['access']))

        if  k['from-epics'] is None:
            validator = None
            
        elif type(k['from-epics']) == type:
            # Instantiate from explicit type
            validator = k['from-epics'](**(kw.get('validator', {})))
            
        elif kw.get('validator') is not None:
            # There's already an instance, but there's an override list
            # of paramters for a re-initialisation available
            validator = k['from-epics'].__class__(**(kw['validator']))
            
        else:
            # If we got here, k['from-epics'] must already be an instance,
            # so we use it.
            validator = k['from-epics']

        setter = PropertySetter(name=suffix, validator=validator, setter=access)

        pv_create_args = k.get('create', {})
        pv_create_args.update(kw.get('create', {}))
        RecordType = k["out"]
        self.pv = RecordType(suffix, always_update=True, on_update=setter, **pv_create_args)
        
        logger.info("Actor %s: kind='%s' actor, access=%r" % (suffix, kw['kind'], access))
    
        
class PropertyConnector(object):
    '''
    Conenctor for a get/set type of property; uses `eda.SignalConnector`
    and `eda.ActorConnector` under the hood to export two PVs, one for
    setting, one for reading back the property. By default the variable
    names end in "VAL" (for setting), respectively "RBV" (for reading).

    A `sync()` method is available to programatically set the
    "VAL" part to its current readback value "RBV".
    '''

    # Using this for validation. For once it's needed in __init__(),
    # and then it's also used in parsing YAML files.
    argnames = { "suffix": str,
                 "name": str,
                 "kind": str,
                 "access": object,
                 "validator": object,
                 "signal": dict,
                 "actor": dict }

    schema = Schema({Optional(k):v for k,v in argnames.items()})

    def __init__(self, ioc_dispatcher, **kw):
        '''
        Parameters:
          - `ioc_dispatcher`: connection to pythonSoftIOC to use.

          - `prefix`: PV prefix, will be prepended to the variable names. By
            default the variable names end in "VAL" (for setting), respectively
            "RBV" (for reading), but they can be overridden by using the `sensor`
            and `actor` parameters. Defaults to `name`.

          - `access`: Access method or (obj, prop) to use for both reading and
            writing. If there are different methods for reading and writing,
            this should be set to `None` and the appropriate methods should
            be set via the `sensor` and `actor` dictionaries.

          - `kind`: The kind of signal to use; must be one of the `eda.ConnectorKinds`
            items.
        
          - `signal`: Dictionary with optional variables to pass on to
            `eda.SignalConnenctor` for the reading part.
        
          - `actor`: Dictionary with optional variables to pass on to
             `eda.ActorConnector` for the setting part.
        '''

        if not set(kw).issubset(set(self.argnames.keys())):
            raise TypeError("Unknown arguments: %r" % list(set(self.argnames.keys()).difference(kw)))

        #print ("Intermediate 1.2: %r" % (kw['access'],))
        
        ad = {}
        ad.update(kw.get('actor', {}))
        ad.setdefault("kind", kw['kind'])

        ad['access'] = self.descend_access(kw['access'], ad.get('access', None))
            

        _s = kw.get('suffix')
        
        ad.setdefault("validator", kw.get('validator', None))
        ad.setdefault("suffix", (_s if _s is not None else kw.get('name'))+"VAL")
        self.val = ActorConnector(ioc_dispatcher, **ad)

        sd = {}
        sd.update(kw.get('signal', {}))
        sd.setdefault("kind", kw['kind'])

        sd['access'] = self.descend_access(kw['access'], sd.get('access', None))
                
        #print ("Intermediate 1.7: %r" % (sd['access'],))
        
        sd.setdefault("validator", kw.get('validator', None))
        sd.setdefault("suffix", (_s if _s is not None else kw.get('name'))+"RBV")
        self.rbv = SignalConnector(ioc_dispatcher, **sd)

        logger.info ("Registering property: %s" % (kw.get('suffix') or kw.get('name')))


    def descend_access(self, top_access, sub_access):
        '''
        Given an existing access object in `top_access` and (possibly)
        a specification for a sub-property access, this returns a merged
        access object to be used.
        '''
        
        if sub_access is None:
            # Nothing to do, there is no sub-access.
            return top_access
            
        if isinstance(sub_access, str):
            # Sub-access is a plain method name (possibly nested).
            # We interpret that to be relative to the top-access object.
            # The top-access object is either directly a Python object,
            # or a (base, "prop") sub-object itself.
            if isinstance(top_access, tuple) \
               and len(top_access)==2 \
               and isinstance(top_access[1], str):
                # top-access is a (base, "prop") notation, so we need
                # to compute the actual property first.
                obj = getattr(*(top_access))
                print("sub-tuple:", obj)
            else:
                # top_access is direct object (callable?), sub-access is just
                # a member nane
                obj = top_access
                print("direct:", obj)
            return FindAccess(obj, sub_access)
        
        elif hasattr(sub_access, "__len__") \
             and len(sub_access) == 2 \
             and isinstance(str, sub_access[1]):
            # Sub-access is a tuple (obj, "attribute"), so we ignore the
            # top-access completely and just build a new access object.
            return FindAccess(*(sub_access['access']))

        # How did we get here?
        raise RuntimeError("Unsupported sub-access type %r for property connector")
    
#
# Validating YAML input for EPICS exports. General YAML format of an export
# record list is like this:
#
#   exports:  # ...or whatever, this isn't regulated (yet)
#
#   - recordType: signal  # indicates essentially "how" this data flows and behaves, can
#                         # be any of the ExportRecordTypes.
#
#     pvName: "PV"  # optional; if not available, then the connector-type entry must have one
#
#     signal:       # depending on the connector type
#
#     - pvName:     # optional; if not available, the one from parent plus a default suffix will be used.
#
#       kind: "analog"  # ...or any other of the ConnectorKinds
#
#       pollPeriod: 0.1 # typical key for "signal" here, other recordTypes may have other fields
#
#
# Schemata for the YAML EPICS-PV "exports" section. General format considerations:
#
#  - `recordType` tells the direction of the data flow and some structural
#    information. Generally there are two basic types -- `signal` (for reading data
#    from external device), and `actor` (for writing data into an external device).
#    Each of these types exports a PV on its own: an OUT, respectively IN semantics
#    of EPICS.
# 
#    A joint type that can do both is `property`: it uses a Signal and an Actor to
#    export two variables, one for reading and one for writing: a "_VAL" and a "_RBV"
#    one.
#
#    Several higher-level types are possible to implement here, for instance a "motor"
#    (usually a positioner / "actor" that has a working state).
#
#
#  - `name`: This is the sine-qua-non. It's the name of the Python object property
#    we're exporting / working on. If the property is a callable, it will be
#    called without an argument (i.e. `obj.prop()`) to read a value, and with
#    one single argument to set a value (`obj.prop(23)`). Otherwise the property
#    will be read and set as if it were a static value (`obj.prop = 23`).
#
#
#  - 'pvSuffix' is an optional suffix for the corresponding EPICS variable;
#    if not set, it defaults to 'name' (with the dots 'n all).
#
#
#  - `format`: Sometimes data comes out of the Python end as a string, containing
#    undesired components (e.g. `"session 1, garble 42.37 yadda"`, and we're
#    actually only interested in `42.37` here). If that's the case with your application,
#    you can specify an explicit format string (here, for instance,
#    `"session {moo}, garble {value} {word}"`) which contains the formatter `{value}`
#    somewhere inside. If that is specified, that will be used for accessing the data.
#
#
#  - `kind` is a hint as to how the data is to be interpreted: it affects two
#    distinct subsystems: on one end, the EPICS PVs that are exported have specific
#    types (aOut/aIn, stringOut/strinIn etc). So specific "kinds" are mapped to
#    specific EPICS PV types.
#    On the other end, the data going in/out, between EPICS/Python (<- which is, to
#    say, the device), also has a may need to be translated to specific Python types.
#    Within Python code, this is done by Validators (somewhat misnamed, they actually
#    also do a great deal of converting).
#
#    Now, with specific `kind` settings come also specific preconfigured validators.
#
#
#  - (FXIME: need to talk about validator settings like `values` etc).
#
#
#  - Reading and writing -- normally reading and writing is done using the same
#    property, in a symmetric manner. For instance: if the property of object `obj`
#    is `prop`, then reading vs. writing is `obj.prop()` vs `obj.prop(...)`, or
#    `obj.prop` vs `obj.prop=...` (if the property is non-callable), or `format(str)`
#    vs. `str.format(...)` if a format is specified, etc. But never, say,
#    `obj.prop()` for reading and `obj.set_prop(42)` for writing.
#
#    Yet sometimes different approaches are necessary.
#
#    For this reason, when using higher-level "recortType" objects like `property`,
#    their lower-level constituents (e.g. `signal` and `actor`), which actually control
#    the reading and writing, mirror the options `name`, `format`, `pvSuffix`.
#    If these are set, they overwrite their conterparts from the supersection.



def FindAccess(obj, name, separators=['.', ':']):
    '''
    Recursively finds a property of object instance `obj`.
    Returns a tuple `(instance, "name")`, where `"name"` is
    the name of the final Python property to be used and
    `instance` is an object instance that owns said property.
    Generally, if `name` is a nested property, then `instance`
    is *not* the same as `obj`.
    '''

    # Pick a separator from the list -- the first one we find.
    for sep in separators:
        if sep in name:
            break
    
    ref = obj
    chain = name.split(sep) if sep in name else [name]
    
    for p in chain[:-1:]:
        ref = getattr(ref, p)
        if ref is None:
            logger.error(f'Property required for "{p}" is None')
            raise RuntimeError(f'Property required for "{p}" is None')
    
    prop = (ref, chain[-1])
    attr = getattr(*prop)
    
    if hasattr(attr, "__call__"):
        return attr
    else:
        return prop


ExportRecordTypes = {
    "signal": SignalConnector,
    "actor": ActorConnector,
    "property": PropertyConnector
}

def ExportObjectFromDict(ioc_dispatch, parent, data,
                         suffixMangle=lambda x: x,
                         recordTypeDefaults=None):
    '''
    Creates an export object (i.e. one of the `ExportObjectTypes`) from
    a data dictionary as specified by the `scm_export_record` and
    kindred schemata. The `parent` is the object to which all properties
    named within `data` relate.
    Parameters:

      - `ioc_dispatch: a pythonSoftIOC dispatch object to use for the
        Signal-type connectors (essentially, those run an asyncio-based
        polling loop, and ioc_dispatch takes care of actually dispatching it)

      - `parent`: The object to which all the properties are refering to

      - `data`: A configuration dictionary cotaining, among others, the
        name of the property to export

      - `suffixMangle`: A callable to optionally do some name mangling on
        the final PV suffix that we'll send to the pythonSoftIOC builder.
        This is typically used e.g. to change case of the variable names,
        or to add more suffixes/prefixes/interfixes as needed.
        This is necessary because we actually want to hide the details of
        the configuration format within `data` to higher layers -- only
        `ExportObjectFromDict()` and functions below it need to know the
        details. Another reason is that details, in particular of variable
        naming, are pretty ofuscated -- naming can be done at high level,
        or may be delegated down to sub-objects, and we don't want higher
        level to have to deal with that. So, in essence: if you need to
        mangle the name, use this function! ;-)
    '''

    # For creating EPICS PV exports delaratively, i.e. from YAML,
    # we start with a 'recordType' (e.g. "signal" or "property"):
    scm_dict = { 'recordType': Or(*(ExportRecordTypes.keys())) }
    
    # The bulk of the parameters may then be specified in sub-dictionaries
    # (see above). Each Connector type class needs to bring its own schema:
    scm_dict.update({ Optional(k):v.schema for k,v in ExportRecordTypes.items() })
    
    # Some record types are primitives ("actor" and "signal"), others are
    # composed ("property"). Things get increasingly nested ("motor" holds
    # several "property" records, for instance). We specify parameters in
    # the order in which they're encapsulated (e.g. "property" accepts dicts
    # for "signal" and "actor" which it's made of).
    # 
    # For convenience, at the top level we also accept parameters common to
    # all subsets (e.g. "name" or "suffix").
    #common_args = reduce(lambda x, y: x&y,
    #                     [set(v.argnames.items()) for k,v in ExportRecordTypes.items()])
    #scm_dict.update({Optional(k):v for k,v in common_args})
    #pprint(scm_dict)
    
    scm = Schema(scm_dict)
    logger.debug("Export schema: %r" % scm)
    data = scm.validate(data)

    # ACHTUNG: This whole passing-arguments-down-to-encapsulated-subtypes shebang
    # only works as long as there is only *one* specific subtype of each being
    # encapsulated, and the corresponding YAML subsection has exactly the same
    # name as the subtype itself. E.g. this works well:
    #
    #  - recordType: property
    #    signal:
    #      ...
    #    actor:
    #      ...
    #
    # This doesn't work:
    #
    # - recordType: motor
    #   signalUpperLimit:
    #     ...
    #   signalLowerLimit:
    #     ...
    #
    
    recType = data['recordType']

    recParams = ((recordTypeDefaults or {}).get(recType) or {}).copy()
    recParams.update(data[recType])

    try:
        recParams['access'] = FindAccess(parent, recParams.get('access') or recParams.get('name'))
    except Exception as e:
        # 'name' may not necessarily be a valid property of the object.
        # So if we fail finding one, we just ignore it, and 'access' will
        # be undefined. It will be up to lower layers to provide a proper
        # access method.
        logger.info('No access for %r defined at top-level' % data.get('name'))
        pass

    # Here's the name mangling. Note that this will only mangle suffix naming
    # at the top level, not at the sub-levels. For that, we're stuck with whatever
    # functionality the Connector objects already implement.
    # The obvious (and possibly the only good) solution to this would be to pass
    # the name-mangling callable down to the Connector types (here: `ExportRecordTypes[]`)
    # constructor itself. For that to work, we'd probably have to derive all Connector
    # classes from a common base class that does handle these types of common tasks
    # (like calling a name mangling lambda).
    recParams['suffix'] = suffixMangle(recParams.get('suffix') or recParams.get('name'))
    
    logger.debug("Building connector %r, settings %r" % (recType, recParams))

    #print ("Intermediate 1: %r" % (recParams['access'],))
    obj = ExportRecordTypes[recType](ioc_dispatch, **recParams)

    return obj



#!/usr/bin/python3

import pytest
from emmi.api import exports

def test_FindAccess():

    class Foo:

        def __init__(self):
            self.moo = True
        
        def func(self):
            return True

        bar = True

        def __call__(self):
            return True

        @property
        def prop(self):
            return True

    foo = Foo()

    assert type(exports.FindAccess(Foo, "bar")) == tuple
    assert getattr(*(exports.FindAccess(Foo, "bar")))

    assert type(exports.FindAccess(Foo, "prop")) == tuple
    assert getattr(*(exports.FindAccess(Foo, "prop")))    

    with pytest.raises(AttributeError):
        assert type(exports.FindAccess(Foo, "moo")) == tuple
        
    assert hasattr(exports.FindAccess(Foo, "func"), "__call__")
    
    assert type(exports.FindAccess(foo, "moo")) == tuple
    assert getattr(*(exports.FindAccess(foo, "moo")))

    assert type(exports.FindAccess(foo, "prop")) == tuple
    assert getattr(*(exports.FindAccess(foo, "prop")))    

    assert type(exports.FindAccess(foo, "bar")) == tuple
    assert getattr(*(exports.FindAccess(foo, "bar")))
    
    assert hasattr(exports.FindAccess(foo, "func"), "__call__")
    assert exports.FindAccess(foo, "func")()



# MappingValidator
def test_MappingValidator():
    
    foo_values={"ON": 1, True: 1, 1: 1,
                "OFF": 0, False: 0, 0: 0}
    
    foo_MappingValidator = exports.MappingValidator(foo_values)
    
    # Test mapping 
    assert foo_MappingValidator["ON"] == 1
    assert foo_MappingValidator["OFF"] == 0
    assert foo_MappingValidator[True] == 1
    assert foo_MappingValidator[False] == 0
    
    # Test mapping error
    with pytest.raises(AssertionError):
        assert foo_MappingValidator["OFF"] == 1

    with pytest.raises(AssertionError):
        assert foo_MappingValidator[False] == 1
    
    with pytest.raises(ValueError):   
        assert foo_MappingValidator["foo"] == 0
        
    with pytest.raises(ValueError, match=": invalid value"): 
        assert foo_MappingValidator["foo"] == 0
    
        
# SetValidator
def test_SetValidator():
    
    foo_values=['foo', 'bar', '1', 0]
    
    foo_SetValidator = exports.SetValidator(values=foo_values)
    
    # Test mapping 
    assert foo_SetValidator['foo'] == 'foo'
    assert foo_SetValidator['bar'] == 'bar'
    assert foo_SetValidator['1'] == '1'
    assert foo_SetValidator[0] == 0
    
    with pytest.raises(ValueError):    
        assert foo_SetValidator['Foo'] == 'Foo'
        
    with pytest.raises(ValueError):    
        assert foo_SetValidator['Foo'] == 'foo'
        
                      
# BoundaryValidator
def test_BoundaryValidator():
    
    foo_limits=[-1, 1]
    foo_inclusiveTT=[True, True]
    foo_inclusiveFT=[False, True]
    foo_inclusiveTF=[True, False]
    foo_inclusiveFF=[False, False]
    
    
    #TT
    foo_BoundaryValidatorTT = exports.BoundaryValidator(limits=foo_limits, inclusive=foo_inclusiveTT, invert=False)
    
    assert foo_BoundaryValidatorTT[-1] == -1
    assert foo_BoundaryValidatorTT[-0.9] == -0.9
    assert foo_BoundaryValidatorTT[1] == 1
    
    with pytest.raises(ValueError):
        assert foo_BoundaryValidatorTT[-1.1] == -1.1
        assert foo_BoundaryValidatorTT[1.1] == 1.1
    
    
    #FT    
    foo_BoundaryValidatorFT = exports.BoundaryValidator(limits=foo_limits, inclusive=foo_inclusiveFT, invert=False)
    
    assert foo_BoundaryValidatorFT[-0.9] == -0.9
    assert foo_BoundaryValidatorFT[1] == 1
    
    with pytest.raises(ValueError):
        assert foo_BoundaryValidatorFT[-1] == -1
        assert foo_BoundaryValidatorFT[1.1] == 1.1
        
        
    #TF    
    foo_BoundaryValidatorTF = exports.BoundaryValidator(limits=foo_limits, inclusive=foo_inclusiveTF, invert=False)
    
    assert foo_BoundaryValidatorTF[-1] == -1
    assert foo_BoundaryValidatorTF[0.9] == 0.9
    
    with pytest.raises(ValueError):
        assert foo_BoundaryValidatorTF[1] == 1
        assert foo_BoundaryValidatorTF[-1.1] == -1.1
        
        
    #FF    
    foo_BoundaryValidatorFF = exports.BoundaryValidator(limits=foo_limits, inclusive=foo_inclusiveFF, invert=False)
    
    assert foo_BoundaryValidatorFF[-0.9] == -0.9
    assert foo_BoundaryValidatorFF[0.9] == 0.9
    
    with pytest.raises(ValueError):
        assert foo_BoundaryValidatorFF[1] == 1
        assert foo_BoundaryValidatorFF[-1] == -1   
        
                
# SliceValidator            
def test_SliceValidator():
    
    foo = ('foofoofoofoofoofoo')
    
    foo_SliceValidator = exports.SliceValidator(start=3, stop=6, step=1)
    assert foo_SliceValidator[foo] == 'foo'
    
    foo_SliceValidator2 = exports.SliceValidator(start=2, stop=8, step=2)
    assert foo_SliceValidator2[foo] == 'oof'
    
    foo_SliceValidator30 = exports.SliceValidator(start=0, stop=30, step=3)
    assert foo_SliceValidator30[foo] == 'ffffff'
    
    
# IdentityValidator
def test_IdentityValidator():
    foo = 2
    bar = 'bar'
    moo = True

    foo_identityValidator = exports.IdentityValidator()
    assert foo_identityValidator[foo] == 2
    assert foo_identityValidator[bar] == 'bar'
    assert foo_identityValidator[moo] == True
    
    
# PropertySetter
# This is difficult - NOT finished!
def test_PropertySetter():
    
    foo_PropertySetter = exports.PropertySetter(name='FOO', setter='foo')

    assert foo_PropertySetter.setter == 'foo'   


#!/usr/bin/python3

import pytest, time, asyncio
from emmi.api.exports import ConnectorKinds, SignalConnector, ActorConnector, \
    PropertyConnector  
    
    
class TestSoftIOC(object):
    '''
    Quick'n Dirty way of setting up an IOC and for testing.
    '''
    def __init__(self, prefix):
        from softioc import softioc, builder, asyncio_dispatcher
        self.builder = builder
        self.dispatcher = asyncio_dispatcher.AsyncioDispatcher()
        self.builder.SetDeviceName(prefix)

        self.exit = softioc.epicsExit

        self.prefix = prefix

        from epics import caget, caput, cainfo
        self.caget = caget
        self.caput = caput
        self.cainfo = cainfo

    def start(self):
        from softioc import softioc
        self.builder.LoadDatabase()
        softioc.iocInit(self.dispatcher)

    def pv(self, name):
        return "%s:%s" % (self.prefix, name)


class ValueAccess(object):
    '''
    Trick class that documents when a getter was called.
    User for unit testing.
    '''
    def __init__(self, val):
        self.tstart = time.time()
        self.count = 0
        self.value = val

    @property
    def age(self):
        return time.time()-self.tstart
        
    def __call__(self, *args):
        self.count += 1
        if len(args) == 0:
            return self.value
        else:
            self.value = args[0]

    def valid(self, mincount=1, age=3.0):
        while self.age < age:
            if self.count >= mincount:
                return True
        return False


@pytest.fixture
def random_value():
    import random
    return random.random()

@pytest.mark.skip(reason="not working yet")
def test_jump_through_hoops(random_value):
    ioc = TestSoftIOC("EMMI_UNITTEST")
    
    a1 = ValueAccess(random_value)
    sen = SignalConnector(ioc.dispatcher, name="signal", kind='analog', access=a1, pollPeriod=0.1)

    a2 = ValueAccess(random_value)
    act = ActorConnector(suffix="actor", access=a2, kind='analog')

    a3 = ValueAccess(0.0)
    prop = PropertyConnector(ioc.dispatcher, name="prop_", kind='analog', access=a3)
    
    ioc.start()
    
    assert a1.valid(mincount=5, age=3.0)

    act.pv.set(3.14)
    time.sleep(1.0)
    assert a2.value == 3.14

    prop.val.pv.set(random_value)
    time.sleep(1.0)
    assert a3.value == random_value




#!/usr/bin/python3

from emmi.api.exports import ConnectorKinds, SignalConnector, ActorConnector, \
    PropertyConnector

import logging, asyncio
logger = logging.getLogger(__name__)

class MotorConnector(object):
    '''
    Implements a simple EPICS motor record, with a small subset of the
    [EPICS motor record](https://epics.anl.gov/bcda/synApps/motor/R7-1/motorRecord.html)
    variables. Relies on a motor inteface as described by `MotorEngine`, i.e.
    a class with the following properties:

      - `position`: returns the current motor values, respectively moves
        to the specified relative value

      - `position_relative`: facilitates a movement relative to the current
        position

      - `state`: a string that begins with one of the following:
    
          - "INIT" (preparing to take commands)
    
          - "IDLE" (not moving, ready to take commands)

          - "STAGING" (about to enter BUSY, but not all BUSY makers are present yet)
    
          - "BUSY" (not able to accept commands, most likely already moving)
    
          - "ERROR" (stopped, reached a situation which requires external action)

          - "STOP" (stopping)
    
          - "FAIL" (unefined state, best effort to be stopped, unrecoverable error)
    
        The state string is intentionally checked only for its beginning part.
        States may be explicitly extended by appending information to strings,
        e.g. "BUSY.TRACKING" or "BUSY.HOMING" to distinguish different flavors
        of states. `SimpleMotor` will not react to anything beyond the five
        1st-level state information, but other software layers may.

    This class needs to be used with a
    [pythonSoftIOC](https://dls-controls.github.io/pythonSoftIOC)
    compatible EPICS IOC Python API. It uses `asyncio` for all asynchronous
    work.

    It implements the EPICS motor record variables VAL, RBV, RVL, HOMF/HOMB,
    STOP.

    FIXME: need to add support for supplementary actions / properties / signals
    (e.g. extend by additional "BUSY" states, or read/write specific properties
    like limits, force probe thresholds etc).
    '''    

    def __init__(self, motor_prefix, motor_engine, ioc_dispatcher,
                 poll_period=0.1, separator="_"):
        '''
        Initializes the IOC variables.

        Args:
            motor_prefix: a string to prepend to the motor variables.
      
            motor_engine: the `MotorEngine` object we'll be talking to.
        
            ioc_dispatcher: a pythonSoftIOC asyncio dispatcher compatible
            instance (typically a `softioc.asyncio_dispatcher.AsyncioDispatcher()`)
        '''

        self.pollPeriod = poll_period

        #from softioc import builder as ioc_builder

        # SPEC needs the following:
        #
        # Can be ignored / business of the controller?
        #  o ACCL: acceleration time in seconds (0.0)
        #  o BDST: backlash distance egu (0)
        #  o BVAL: backlash velocity egu/s (0)
        #  o VBAS: base velocity (minimum velocity?) egu/s (0)
        #  o ERES: encoder step size egu (0)
        #  o MRES: motor step size egu (0)
        #
        # Calibration fields and coordinate system transformations:
        #  - SET: set/use switch for calibration fields (0: use, 1: set)
        #  - FOFF: offset freeze switch -- is the user prevented from
        #          writing the offset?
        #  - OFF: user offset egu
        #  + DIR: user direction        
        #  - RRBV: raw readback value
        #  - RVAL: raw desired value        
        #
        # Unclear:
        #  o UEIP: use encoder if present (always 1?)
        #  o VELO: velocity egu/s (set to 0?)
        #
        # Need to have / already have:
        #  o STOP: stop
        #  o VAL: user desired value
        #  - SPMG: stop/pause/move/go -- complicated beast
        #    - Stop: same as STOP?        
        #
        # Nice to have, but not part of the EDA Motor Model:
        #  + DHLM: dial high-limit
        #  + DLLM: dial low-limit
        #  + HLS: at high-limit switch
        #  + LLS: at low-limit switch        
        #  o DMOV: done moving to value
        #
        # Unknown:
        #  + DISP: disable (turn off motor/unusable)        
        #
        # Nice to have, not needed by SPEC:
        #  o EGU: engineering unit names
        #  - RLV: relative-move value: when changed, changes VAL,
        #    then resets itself to 0
        #  
        
        self.engine = motor_engine

        # VAL/RBV
        self.con_pos = PropertyConnector(ioc_dispatcher,
                                         suffix=motor_prefix+separator,
                                         access=self.motor_position,
                                         signal={ 'pollPeriod': poll_period },
                                         kind="analog")

        # STATE (motor engine)
        self.con_state = SignalConnector(ioc_dispatcher,
                                         suffix=motor_prefix+separator+"STATE",
                                         access=(motor_engine, "state"),
                                         pollPeriod=poll_period,
                                         kind="strings",
                                         validator={ 'values': [
                                             'INIT', 'IDLE', 'STAGING', 'BUSY',
                                             'STOP', 'ERROR', 'FATAL'
                                         ]})

        # STOP
        self.con_stop = ActorConnector(ioc_dispatch=ioc_dispatcher,
                                       suffix=motor_prefix+separator+"STOP",
                                       access=self.conExecStop,
                                       kind="values",
                                       validator={'values': [0, 1]})

        self.motor_name = motor_prefix
        self.motor_prefix = motor_prefix
        self.prefix_separator = separator

        # for use with .addBusy()
        self.additional_actions = {}

        self.ioc_dispatcher = ioc_dispatcher


    def motor_position(self, val=None):
        # Use this to query and/or set (user) motor position.
        # CalibrationMixin will properly overwrite this.
        if val is None:
            return self.engine.position
        else:
            self.engine.position = val


    def addBusy(self, action_suffix, go, **go_params):
        '''
        Adds an additional motor work trigger with the specified suffix.
        
        Some motors support additional actions in their BUSY phase (e.g.
        relative move, homing action etc). This supports a simple extension
        mechanism which takes a EPICS variable suffix (e.g. "HOMF") and
        a set of parameters for the `MotorEngine.go()` function.
        '''

        class MarkGoUnmark:
            def __init__(self, go_name, go_params, engine, marker_pv):
                self.go_name = go_name
                self.go_params = go_params
                self.engine = engine
                self.marker_pv = marker_pv
            
            # wrapper to mark
            async def __call__(self, val):

                if val == 0:
                    logger.debug(f'{self.go_name}: val={val} requested, nothing to do')
                    return

                logger.debug(f'{self.go_name}: triggered, val={val}')
                logger.debug(f"Requesting BUSY/{self.go_name}")

                # triggering action
                self.engine.go(self.go_name, **self.go_params)

                # ...wait for action to start...
                while self.engine.state != 'BUSY':
                    await asyncio.sleep(0.1)

                # ...then wait for action to finish
                while True:
                    if (self.engine.state != "BUSY") or \
                       (self.engine.state == "BUSY") and (self.engine.business != self.go_name):
                        logger.debug(f"State: {self.engine.state}, business: {self.engine.business}")
                        logger.info(f"Done with BUSY/{self.go_name}")
                        break
                    await asyncio.sleep(0.1)
                
                try:
                    logger.debug(f'{self.go_name}: done, removing mark')
                    self.marker_pv.pv.set(0)

                except Exception as e:
                    # this happens on Raspi
                    logger.error(f'{self.go_name}: done, but val=0 update failed')


        thing_doer = MarkGoUnmark(go, go_params, self.engine, None)
        self.additional_actions[action_suffix] = \
            ActorConnector(self.ioc_dispatcher,
                           suffix=self.motor_prefix+self.prefix_separator+action_suffix,
                           access=thing_doer,
                           kind="integer")
        
        thing_doer.marker_pv = self.additional_actions[action_suffix]


    async def conExecStop(self, val):
        '''
        EPICS motor STOP command: if val==1, it executes the STOP command,
        then it sets VAL to current position and resets itself to 0.
        '''
        if val != 1:
            return

        self.engine.stop()
        self.con_stop.pv.set(0)
        
        while self.engine.state != "IDLE":
            await asyncio.sleep(self.pollPeriod)

        logger.info(f'Overriding VAL to display current value ({self.engine.position})')
        self.con_pos.val.pv.set(self.motor_position())


class DMOVMixin:
    ''' Implements the DMOV motor suffix (returns 1 while motor is not IDLE).
    '''
    def __init__(self):
        
        # DMOV - set to 0 when motor begins moving
        self.con_dmov = SignalConnector(self.ioc_dispatcher,
                                        suffix=self.motor_prefix+self.prefix_separator+"DMOV",
                                        access=lambda: int(self.engine.state == "IDLE"),
                                        kind="integer",
                                        pollPeriod=self.pollPeriod)


class MSTAMixin:
    ''' Implements a SPEC-like STATUS motor suffix (returns 0x02 bit set while moving).

    EPICS (motor record) actually defines the following bits:
      1. DIRECTION: last raw direction; (0:Negative, 1:Positive)
      2. DONE: motion is complete.
      3. PLUS_LS: plus limit switch has been hit.
      4. HOMELS: state of the home limit switch.
      5. Unused
      6. POSITION: closed-loop position control is enabled.
      7. SLIP_STALL: Slip/Stall detected (eg. fatal following error)
      8. HOME: if at home position.
      9. PRESENT: encoder is present.
     10. PROBLEM: driver stopped polling, or hardware problem
     11. MOVING: non-zero velocity present.
     12. GAIN_SUPPORT: motor supports closed-loop position control.
     13. COMM_ERR: Controller communication error.
     14. MINUS_LS: minus limit switch has been hit.
     15. HOMED: the motor has been homed.
    '''

    def __init__(self):
        
        suffix=self.motor_prefix+self.prefix_separator+"MSTA"

        status_lambda = lambda: 0x02 if not (self.engine.state in ("IDLE", )) else 0x00
        
        self.con_status = SignalConnector(self.ioc_dispatcher,
                                          suffix=suffix,
                                          access=status_lambda,
                                          kind="integer",
                                          pollPeriod=self.pollPeriod)


class CalibrationMixin:
    '''
    Adds calibration and user/dial coordinate transformation functionality.
    If you add this to your motor, the `.position` property will, in fact,
    return "user" coordinates, which differ from the "dial" coordinates
    (i.e. the raw output of `Motor.where()`) by the equation
    `user = dial*dir + offs`.
    
    These will be exposed via `VAL` / `RBV`.

    The raw output of `.where()` will actually be exposed
    in `DVAL` / `DRBV` / `DMOV`.

    Additional variables will be added:
     - `DIR` for direction
     - `OFF` for offset
     - `FOFF` to freeze / unfreeze offset (always `0`, i.e. "variable")
     - `SET` / `SSET` / `SUSE`
    '''
    
    def __init__(self, offset=0, direction=1):
        self._usr_offset = offset
        self._usr_direction = direction

        self.con_dial_pos = PropertyConnector(self.ioc_dispatcher,
                                              suffix=f'{self.motor_prefix}{self.prefix_separator}D',
                                              access=self.dial_motor_position,
                                              signal={ 'pollPeriod': self.pollPeriod },
                                              kind="analog")


        self.con_offs = ActorConnector(ioc_dispatch=self.ioc_dispatcher,
                                       suffix=self.motor_prefix+self.prefix_separator+"OFF",
                                       access=self.user_offset,
                                       kind="analog")


        self.con_dir = ActorConnector(ioc_dispatch=self.ioc_dispatcher,
                                      suffix=self.motor_prefix+self.prefix_separator+"DIR",
                                      access=self.user_direction,
                                      kind="analog")

        # EMMI doesn't (yet) explicitly support "r/w actors". Actors are write-only (from
        # the EPICS user's perspective).
        # But there's nothing wrong with setting a PV from this side of the code, we just
        # have to do it explicitly. And in particular, we have to do this for the
        # initialization.
        self.con_offs.pv.set(self._usr_offset)
        self.con_dir.pv.set(self._usr_direction)
        

    def user_offset(self, val=None):
        # Used primarily to set the user offset from the EPICS side, through PVs
        if val is None:
            return self._usr_offset
        else:
            self._usr_offset = val


    def user_direction(self, val=None):
        # Used to set direction (0=pos, 1=neg), to EPICS conventions.
        # Returns the mathematical direction (i.e. -1 / +1, depending
        # on the ._usr_direction field)
        if val is None:
            return { 0: 1, 1: -1}[self._usr_direction]
        else:
            self._usr_direction = val


    def dial_motor_position(self, val=None):
        if val is None:
            return self.engine.position
        else:
            self.engine.position = val


    def user_motor_position(self, val=None):
        if val is None:
            return self.dial_motor_position()*self.user_direction() + self._usr_offset
        else:
            self.dial_motor_position((val-self._usr_offset)/self.user_direction())


    def motor_position(self, val=None):
        # Hijacking the .motor_position() of the parent class to
        # actually display the user-position instead of the original
        # hardware position
        return self.user_motor_position(val)


    async def conExecStop(self, val):
        await super().conExecStop(val)
        
        if val != 1:
            return

        self.con_dial_pos.val.pv.set(self.dial_motor_position())


class ErrorMixin:
    ''' Adds error read / clear functionality '''
    def __init__(self):
        self.con_eclr = ActorConnector(ioc_dispatch=self.ioc_dispatcher,
                                       suffix=self.motor_prefix+self.prefix_separator+"ECLR",
                                       access=self.clear_motor,
                                       kind="values",
                                       validator={'values': [0, 1]})

    def clear_motor(self, val):
        if val > 0:
            self.engine.clear_motor()
            self.con_eclr.pv.set(0)
        
        
    
        
class SpecMixin:
    ''' Adds a number of nonsensical fields to keep Spec happy '''

    def __init__(self):
        
        # lots of variables expected by spec but which we don't really serve
        self.con_constants = [
            SignalConnector(self.ioc_dispatcher,
                            suffix=self.motor_prefix+self.prefix_separator+s,
                            access=lambda: value,
                            kind=kind,
                            pollPeriod=10.0)
            for s,kind,value in [ ("ACCL", "analog", 0),
                                  ("BDST", "analog", 0),
                                  ("BVAL", "analog", 0),
                                  ("VBAS", "analog", 0),
                                  ("ERES", "analog", 0),
                                  ("MRES", "analog", 0),
                                  ("UEIP", "analog", 1),
                                  ("VELO", "analog", 0)]]
        #("EGU",  "text", "mm")] ]
